import datetime
import importlib
import importlib.metadata
import os
import re
from typing import Literal

from bec_qthemes import material_icon
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QInputDialog, QMessageBox, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty
from bec_widgets.widgets.containers.explorer.collapsible_tree_section import CollapsibleSection
from bec_widgets.widgets.containers.explorer.explorer import Explorer
from bec_widgets.widgets.containers.explorer.macro_tree_widget import MacroTreeWidget
from bec_widgets.widgets.containers.explorer.script_tree_widget import ScriptTreeWidget


class IDEExplorer(BECWidget, QWidget):
    """Integrated Development Environment Explorer"""

    PLUGIN = True
    RPC = False

    file_open_requested = Signal(str, str)
    file_preview_requested = Signal(str, str)

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self._sections = []  # Use list to maintain order instead of set
        self.main_explorer = Explorer(parent=self)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.main_explorer)
        self.setLayout(layout)
        self.sections = ["scripts", "macros"]

    @SafeProperty(list)
    def sections(self):
        return list(self._sections)

    @sections.setter
    def sections(self, value):
        existing_sections = set(self._sections)
        new_sections = set(value)
        # Find sections to add, maintaining the order from the input value list
        sections_to_add = [
            section for section in value if section in (new_sections - existing_sections)
        ]
        self._sections = list(value)  # Store as ordered list
        self._update_section_visibility(sections_to_add)

    def _update_section_visibility(self, sections):
        # sections is now an ordered list, not a set
        for section in sections:
            self._add_section(section)

    def _add_section(self, section_name):
        match section_name.lower():
            case "scripts":
                self.add_script_section()
            case "macros":
                self.add_macro_section()
            case _:
                pass

    def _remove_section(self, section_name):
        section = self.main_explorer.get_section(section_name.upper())
        if section:
            self.main_explorer.remove_section(section)
            self._sections.remove(section_name)

    def clear(self):
        """Clear all sections from the explorer."""
        for section in reversed(self._sections):
            self._remove_section(section)

    def add_script_section(self):
        section = CollapsibleSection(parent=self, title="SCRIPTS", indentation=0)

        script_explorer = Explorer(parent=self)
        script_widget = ScriptTreeWidget(parent=self)
        script_widget.file_open_requested.connect(self._emit_file_open_scripts_local)
        script_widget.file_selected.connect(self._emit_file_preview_scripts_local)
        local_scripts_section = CollapsibleSection(title="Local", show_add_button=True, parent=self)
        local_scripts_section.header_add_button.clicked.connect(self._add_local_script)
        local_scripts_section.set_widget(script_widget)
        local_script_dir = self.client._service_config.model.user_scripts.base_path
        if not os.path.exists(local_script_dir):
            os.makedirs(local_script_dir)
        script_widget.set_directory(local_script_dir)
        script_explorer.add_section(local_scripts_section)

        section.set_widget(script_explorer)
        self.main_explorer.add_section(section)

        plugin_scripts_dir = self._get_plugin_dir("scripts")

        if not plugin_scripts_dir or not os.path.exists(plugin_scripts_dir):
            return
        shared_script_section = CollapsibleSection(title="Shared (Read-only)", parent=self)
        shared_script_section.setToolTip("Shared scripts (read-only)")
        shared_script_widget = ScriptTreeWidget(parent=self)
        shared_script_section.set_widget(shared_script_widget)
        shared_script_widget.set_directory(plugin_scripts_dir)
        script_explorer.add_section(shared_script_section)
        shared_script_widget.file_open_requested.connect(self._emit_file_open_scripts_shared)
        shared_script_widget.file_selected.connect(self._emit_file_preview_scripts_shared)

    def add_macro_section(self):
        section = CollapsibleSection(
            parent=self,
            title="MACROS",
            indentation=0,
            show_add_button=True,
            tooltip="Macros are reusable functions that can be called from scripts or the console.",
        )
        section.header_add_button.setIcon(
            material_icon("refresh", size=(20, 20), convert_to_pixmap=False)
        )
        section.header_add_button.setToolTip("Reload all macros")
        section.header_add_button.clicked.connect(self._reload_macros)

        macro_explorer = Explorer(parent=self)
        macro_widget = MacroTreeWidget(parent=self)
        macro_widget.macro_open_requested.connect(self._emit_file_open_macros_local)
        macro_widget.macro_selected.connect(self._emit_file_preview_macros_local)
        local_macros_section = CollapsibleSection(title="Local", show_add_button=True, parent=self)
        local_macros_section.header_add_button.clicked.connect(self._add_local_macro)
        local_macros_section.set_widget(macro_widget)
        local_macro_dir = self.client._service_config.model.user_macros.base_path
        if not os.path.exists(local_macro_dir):
            os.makedirs(local_macro_dir)
        macro_widget.set_directory(local_macro_dir)
        macro_explorer.add_section(local_macros_section)

        section.set_widget(macro_explorer)
        self.main_explorer.add_section(section)

        plugin_macros_dir = self._get_plugin_dir("macros")

        if not plugin_macros_dir or not os.path.exists(plugin_macros_dir):
            return
        shared_macro_section = CollapsibleSection(title="Shared (Read-only)", parent=self)
        shared_macro_section.setToolTip("Shared macros (read-only)")
        shared_macro_widget = MacroTreeWidget(parent=self)
        shared_macro_section.set_widget(shared_macro_widget)
        shared_macro_widget.set_directory(plugin_macros_dir)
        macro_explorer.add_section(shared_macro_section)
        shared_macro_widget.macro_open_requested.connect(self._emit_file_open_macros_shared)
        shared_macro_widget.macro_selected.connect(self._emit_file_preview_macros_shared)

    def _get_plugin_dir(self, dir_name: Literal["scripts", "macros"]) -> str | None:
        """Get the path to the specified directory within the BEC plugin.

        Returns:
            The path to the specified directory, or None if not found.
        """
        plugins = importlib.metadata.entry_points(group="bec")
        for plugin in plugins:
            if plugin.name == "plugin_bec":
                plugin = plugin.load()
                return os.path.join(plugin.__path__[0], dir_name)
        return None

    def _emit_file_open_scripts_local(self, file_name: str):
        self.file_open_requested.emit(file_name, "scripts/local")

    def _emit_file_preview_scripts_local(self, file_name: str):
        self.file_preview_requested.emit(file_name, "scripts/local")

    def _emit_file_open_scripts_shared(self, file_name: str):
        self.file_open_requested.emit(file_name, "scripts/shared")

    def _emit_file_preview_scripts_shared(self, file_name: str):
        self.file_preview_requested.emit(file_name, "scripts/shared")

    def _emit_file_open_macros_local(self, function_name: str, file_path: str):
        self.file_open_requested.emit(file_path, "macros/local")

    def _emit_file_preview_macros_local(self, function_name: str, file_path: str):
        self.file_preview_requested.emit(file_path, "macros/local")

    def _emit_file_open_macros_shared(self, function_name: str, file_path: str):
        self.file_open_requested.emit(file_path, "macros/shared")

    def _emit_file_preview_macros_shared(self, function_name: str, file_path: str):
        self.file_preview_requested.emit(file_path, "macros/shared")

    def _add_local_script(self):
        """Show a dialog to enter the name of a new script and create it."""

        target_section = self.main_explorer.get_section("SCRIPTS")
        script_dir_section = target_section.content_widget.get_section("Local")

        local_script_dir = script_dir_section.content_widget.directory

        # Prompt user for filename
        filename, ok = QInputDialog.getText(
            self, "New Script", f"Enter script name ({local_script_dir}/<filename>):"
        )

        if not ok or not filename:
            return  # User cancelled or didn't enter a name

        # Add .py extension if not already present
        if not filename.endswith(".py"):
            filename = f"{filename}.py"

        file_path = os.path.join(local_script_dir, filename)

        # Check if file already exists
        if os.path.exists(file_path):
            response = QMessageBox.question(
                self,
                "File exists",
                f"The file '{filename}' already exists. Do you want to overwrite it?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )

            if response != QMessageBox.StandardButton.Yes:
                return  # User chose not to overwrite

        try:
            # Create the file with a basic template
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(f"""
\"\"\"
{filename} - Created at {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}
\"\"\"
""")

        except Exception as e:
            # Show error if file creation failed
            QMessageBox.critical(self, "Error", f"Failed to create script: {str(e)}")

    def _add_local_macro(self):
        """Show a dialog to enter the name of a new macro function and create it."""

        target_section = self.main_explorer.get_section("MACROS")
        macro_dir_section = target_section.content_widget.get_section("Local")

        local_macro_dir = macro_dir_section.content_widget.directory

        # Prompt user for function name
        function_name, ok = QInputDialog.getText(self, "New Macro", f"Enter macro function name:")

        if not ok or not function_name:
            return  # User cancelled or didn't enter a name

        # Sanitize function name
        function_name = re.sub(r"[^a-zA-Z0-9_]", "_", function_name)
        if not function_name or function_name[0].isdigit():
            QMessageBox.warning(
                self, "Invalid Name", "Function name must be a valid Python identifier."
            )
            return

        # Create filename based on function name
        filename = f"{function_name}.py"
        file_path = os.path.join(local_macro_dir, filename)

        # Check if file already exists
        if os.path.exists(file_path):
            response = QMessageBox.question(
                self,
                "File exists",
                f"The file '{filename}' already exists. Do you want to overwrite it?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )

            if response != QMessageBox.StandardButton.Yes:
                return  # User chose not to overwrite

        try:
            # Create the file with a macro function template
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(f'''"""
{function_name} macro - Created at {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}
"""


def {function_name}():
    """
    Description of what this macro does.
    
    Add your macro implementation here.
    """
    print("Executing macro: {function_name}")
    # TODO: Add your macro code here
    pass
''')

            # Refresh the macro tree to show the new function
            macro_dir_section.content_widget.refresh()

        except Exception as e:
            # Show error if file creation failed
            QMessageBox.critical(self, "Error", f"Failed to create macro: {str(e)}")

    def _reload_macros(self):
        """Reload all macros using the BEC client."""
        try:
            if hasattr(self.client, "macros"):
                self.client.macros.load_all_user_macros()

                # Refresh the macro tree widgets to show updated functions
                target_section = self.main_explorer.get_section("MACROS")
                if target_section and hasattr(target_section, "content_widget"):
                    local_section = target_section.content_widget.get_section("Local")
                    if local_section and hasattr(local_section, "content_widget"):
                        local_section.content_widget.refresh()

                    shared_section = target_section.content_widget.get_section("Shared")
                    if shared_section and hasattr(shared_section, "content_widget"):
                        shared_section.content_widget.refresh()

                QMessageBox.information(
                    self, "Reload Macros", "Macros have been reloaded successfully."
                )
            else:
                QMessageBox.warning(self, "Reload Macros", "Macros functionality is not available.")
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to reload macros: {str(e)}")

    def refresh_macro_file(self, file_path: str):
        """Refresh a single macro file in the tree widget.

        Args:
            file_path: Path to the macro file that was updated
        """
        target_section = self.main_explorer.get_section("MACROS")
        if not target_section or not hasattr(target_section, "content_widget"):
            return

        # Determine if this is a local or shared macro based on the file path
        local_section = target_section.content_widget.get_section("Local")
        shared_section = target_section.content_widget.get_section("Shared")

        # Check if file belongs to local macros directory
        if (
            local_section
            and hasattr(local_section, "content_widget")
            and hasattr(local_section.content_widget, "directory")
        ):
            local_macro_dir = local_section.content_widget.directory
            if local_macro_dir and file_path.startswith(local_macro_dir):
                local_section.content_widget.refresh_file_item(file_path)
                return

        # Check if file belongs to shared macros directory
        if (
            shared_section
            and hasattr(shared_section, "content_widget")
            and hasattr(shared_section.content_widget, "directory")
        ):
            shared_macro_dir = shared_section.content_widget.directory
            if shared_macro_dir and file_path.startswith(shared_macro_dir):
                shared_section.content_widget.refresh_file_item(file_path)
                return


if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication

    app = QApplication([])
    script_explorer = IDEExplorer()
    script_explorer.show()
    app.exec_()



from __future__ import annotations

from qtpy.QtCore import QEasingCurve, QParallelAnimationGroup, QPropertyAnimation
from qtpy.QtWidgets import QGraphicsOpacityEffect, QWidget

ANIMATION_DURATION = 500  # ms


class RevealAnimator:
    """Animate reveal/hide for a single widget using opacity + max W/H.

    This keeps the widget always visible to avoid jitter from setVisible().
    Collapsed state: opacity=0, maxW=0, maxH=0.
    Expanded state: opacity=1, maxW=sizeHint.width(), maxH=sizeHint.height().
    """

    def __init__(
        self,
        widget: QWidget,
        duration: int = ANIMATION_DURATION,
        easing: QEasingCurve.Type = QEasingCurve.InOutCubic,
        initially_revealed: bool = False,
        *,
        animate_opacity: bool = True,
        animate_width: bool = True,
        animate_height: bool = True,
    ):
        self.widget = widget
        self.animate_opacity = animate_opacity
        self.animate_width = animate_width
        self.animate_height = animate_height
        # Opacity effect
        self.fx = QGraphicsOpacityEffect(widget)
        widget.setGraphicsEffect(self.fx)
        # Animations
        self.opacity_anim = (
            QPropertyAnimation(self.fx, b"opacity") if self.animate_opacity else None
        )
        self.width_anim = (
            QPropertyAnimation(widget, b"maximumWidth") if self.animate_width else None
        )
        self.height_anim = (
            QPropertyAnimation(widget, b"maximumHeight") if self.animate_height else None
        )
        for anim in (self.opacity_anim, self.width_anim, self.height_anim):
            if anim is not None:
                anim.setDuration(duration)
                anim.setEasingCurve(easing)
        # Initialize to requested state
        self.set_immediate(initially_revealed)

    def _natural_sizes(self) -> tuple[int, int]:
        sh = self.widget.sizeHint()
        w = max(sh.width(), 1)
        h = max(sh.height(), 1)
        return w, h

    def set_immediate(self, revealed: bool):
        """
        Immediately set the widget to the target revealed/collapsed state.

        Args:
            revealed(bool): True to reveal, False to collapse.
        """
        w, h = self._natural_sizes()
        if self.animate_opacity:
            self.fx.setOpacity(1.0 if revealed else 0.0)
        if self.animate_width:
            self.widget.setMaximumWidth(w if revealed else 0)
        if self.animate_height:
            self.widget.setMaximumHeight(h if revealed else 0)

    def setup(self, reveal: bool):
        """
        Prepare animations to transition to the target revealed/collapsed state.

        Args:
            reveal(bool): True to reveal, False to collapse.
        """
        # Prepare animations from current state to target
        target_w, target_h = self._natural_sizes()
        if self.opacity_anim is not None:
            self.opacity_anim.setStartValue(self.fx.opacity())
            self.opacity_anim.setEndValue(1.0 if reveal else 0.0)
        if self.width_anim is not None:
            self.width_anim.setStartValue(self.widget.maximumWidth())
            self.width_anim.setEndValue(target_w if reveal else 0)
        if self.height_anim is not None:
            self.height_anim.setStartValue(self.widget.maximumHeight())
            self.height_anim.setEndValue(target_h if reveal else 0)

    def add_to_group(self, group: QParallelAnimationGroup):
        """
        Add the prepared animations to the given animation group.

        Args:
            group(QParallelAnimationGroup): The animation group to add to.
        """
        if self.opacity_anim is not None:
            group.addAnimation(self.opacity_anim)
        if self.width_anim is not None:
            group.addAnimation(self.width_anim)
        if self.height_anim is not None:
            group.addAnimation(self.height_anim)

    def animations(self):
        """
        Get a list of all animations (non-None) for adding to a group.
        """
        return [
            anim
            for anim in (self.opacity_anim, self.height_anim, self.width_anim)
            if anim is not None
        ]



"""Dialogs for device configuration forms and ophyd testing."""

from typing import Any, Iterable, Tuple

from bec_lib.atlas_models import Device as DeviceModel
from bec_lib.logger import bec_logger
from ophyd_devices.interfaces.device_config_templates.ophyd_templates import OPHYD_DEVICE_TEMPLATES
from qtpy import QtCore, QtWidgets

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.widgets.control.device_manager.components import OphydValidation
from bec_widgets.widgets.control.device_manager.components.device_config_template.device_config_template import (
    DeviceConfigTemplate,
)
from bec_widgets.widgets.control.device_manager.components.device_config_template.template_items import (
    validate_name,
)
from bec_widgets.widgets.control.device_manager.components.ophyd_validation import (
    ConfigStatus,
    ConnectionStatus,
    format_error_to_md,
)

DEFAULT_DEVICE = "CustomDevice"
_ValidationResultIter = Iterable[Tuple[dict[str, Any], ConfigStatus, ConnectionStatus, str]]


logger = bec_logger.logger


class DeviceManagerOphydValidationDialog(QtWidgets.QDialog):
    """Popup dialog to test Ophyd device configurations interactively."""

    def __init__(self, parent=None, config: dict | None = None):  # type: ignore
        super().__init__(parent)
        self.setWindowTitle("Device Manager Ophyd Test")
        self._config_status = ConfigStatus.UNKNOWN.value
        self._connection_status = ConnectionStatus.UNKNOWN.value
        self._validated_config: dict = {}
        self._validation_msg: str = ""

        layout = QtWidgets.QVBoxLayout(self)

        # Core test widget
        self.device_manager_ophyd_test = OphydValidation()
        layout.addWidget(self.device_manager_ophyd_test)

        # Log/Markdown box for messages
        self.text_box = QtWidgets.QTextEdit()
        self.text_box.setReadOnly(True)
        layout.addWidget(self.text_box)

        # Load and apply configuration
        config = config or {}
        device_name = config.get("name", None)
        if device_name:
            self.device_manager_ophyd_test.add_device_to_keep_visible_after_validation(device_name)

        # Dialog Buttons: equal size, stacked horizontally
        button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.StandardButton.Close)
        for button in button_box.buttons():
            button.setSizePolicy(
                QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
            )
            button.clicked.connect(self.accept)
        # button_box.setCenterButtons(False)
        layout.addWidget(button_box)
        self.device_manager_ophyd_test.validation_completed.connect(self._on_device_validated)
        self._resize_dialog()
        self.finished.connect(self._finished)

        # Add and test device config
        self.device_manager_ophyd_test.change_device_configs([config], added=True, connect=True)

    def _resize_dialog(self):
        """Resize the dialog based on the screen size."""
        app: QtCore.QCoreApplication = QtWidgets.QApplication.instance()
        screen = app.primaryScreen()
        screen_geometry = screen.availableGeometry()
        screen_width = screen_geometry.width()
        screen_height = screen_geometry.height()
        # 70% of screen height, keep 4:3 ratio
        height = int(screen_height * 0.7)
        width = int(height * (4 / 3))

        # If width exceeds screen width, scale down
        if width > screen_width * 0.9:
            width = int(screen_width * 0.9)
            height = int(width / (4 / 3))

        self.resize(width, height)

    def _on_device_validated(
        self, device_config: dict, config_status: int, connection_status: int, validation_msg: str
    ):
        device_name = device_config.get("name", "")
        self._config_status = config_status
        self._connection_status = connection_status
        self._validated_config = device_config
        self._validation_msg = validation_msg
        self.text_box.setMarkdown(format_error_to_md(device_name, validation_msg))

    @SafeSlot(int)
    def _finished(self, state: int):
        self.device_manager_ophyd_test.close()
        self.device_manager_ophyd_test.deleteLater()

    @property
    def validation_result(self) -> tuple[dict, int, int, str]:
        """
        Return the result of the validation as a tuple of

        Returns:
            result (Tuple[dict, int, int]): A tuple containing:
                                            validated_config (dict): The validated device configuration.
                                            config_status (int): The configuration status.
                                            connection_status (int): The connection status.

        """
        return (
            self._validated_config,
            self._config_status,
            self._connection_status,
            self._validation_msg,
        )


class DeviceFormDialog(QtWidgets.QDialog):

    # Signal emitted when device configuration is accepted, only
    # emitted when the user clicks the "Add Device" button
    # The integer values indicate if the device config was
    # validated: config_status, connection_status
    accepted_data = QtCore.Signal(dict, int, int, str, str)

    def __init__(self, parent=None, add_btn_text: str = "Add Device"):  # type: ignore
        super().__init__(parent)
        # Track old device name if config is edited
        self._old_device_name: str = ""

        # Config validation result
        self._validation_result: tuple[dict, int, int, str] = (
            {},
            ConfigStatus.UNKNOWN.value,
            ConnectionStatus.UNKNOWN.value,
            "",
        )
        # Group to variants mapping
        self._group_variants: dict[str, list[str]] = {
            group: [variant for variant in variants.keys()]
            for group, variants in OPHYD_DEVICE_TEMPLATES.items()
        }

        self._control_widgets: dict[str, QtWidgets.QWidget] = {}

        # Setup layout
        self.setWindowTitle("Device Config Dialog")
        layout = QtWidgets.QVBoxLayout(self)

        # Control panel
        self._control_box = self.create_control_panel()
        layout.addWidget(self._control_box)

        # Device config template display
        self._device_config_template = DeviceConfigTemplate(parent=self)
        self._frame = QtWidgets.QFrame()
        self._frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self._frame.setFrameShadow(QtWidgets.QFrame.Raised)
        frame_layout = QtWidgets.QVBoxLayout(self._frame)
        frame_layout.addWidget(self._device_config_template)
        layout.addWidget(self._frame)

        # Custom buttons
        self.add_btn = QtWidgets.QPushButton(add_btn_text)
        self.test_connection_btn = QtWidgets.QPushButton("Test Connection")
        self.cancel_btn = QtWidgets.QPushButton("Cancel")
        self.reset_btn = QtWidgets.QPushButton("Reset Form")

        btn_box = QtWidgets.QDialogButtonBox(self)
        btn_box.addButton(self.cancel_btn, QtWidgets.QDialogButtonBox.ButtonRole.RejectRole)
        btn_box.addButton(self.reset_btn, QtWidgets.QDialogButtonBox.ButtonRole.ActionRole)
        btn_box.addButton(
            self.test_connection_btn, QtWidgets.QDialogButtonBox.ButtonRole.ActionRole
        )
        btn_box.addButton(self.add_btn, QtWidgets.QDialogButtonBox.ButtonRole.AcceptRole)
        for btn in btn_box.buttons():
            btn.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        layout.addWidget(btn_box)

        frame_layout.addWidget(btn_box)

        # Connect signals to explicit slots
        self.add_btn.clicked.connect(self._add_config)
        self.test_connection_btn.clicked.connect(self._test_connection)
        self.reset_btn.clicked.connect(self._reset_config)
        self.cancel_btn.clicked.connect(self._reject_config)

        # layout.addWidget(self._device_config_template)
        self.update_variant_combo(self._control_widgets["group_combo"].currentText())
        self.finished.connect(self._finished)

        # Wait dialog when adding config
        self._wait_dialog: QtWidgets.QProgressDialog | None = None

    @SafeSlot(int)
    def _finished(self, state: int):
        for widget in self._control_widgets.values():
            widget.close()
            widget.deleteLater()
        if self._wait_dialog is not None:
            self._wait_dialog.close()
            self._wait_dialog.deleteLater()

    @property
    def config_validation_result(self) -> tuple[dict, int, int, str]:
        """Return the result of the last configuration validation."""
        return self._validation_result

    @config_validation_result.setter
    def config_validation_result(self, result: tuple[dict, int, int, str]):
        self._validation_result = result

    def set_device_config(self, device_config: dict):
        """Set the device configuration in the template form."""
        # Figure out which group and variant this config belongs to
        device_class = device_config.get("deviceClass", None)
        for group, variants in OPHYD_DEVICE_TEMPLATES.items():
            for variant, template_info in variants.items():
                if template_info.get("deviceClass", None) == device_class:
                    # Found the matching group and variant
                    self._control_widgets["group_combo"].setCurrentText(group)
                    self.update_variant_combo(group)
                    self._control_widgets["variant_combo"].setCurrentText(variant)
                    self._device_config_template.set_config_fields(device_config)
                    return
        # If no match found, set to default
        self._control_widgets["group_combo"].setCurrentText(DEFAULT_DEVICE)
        self.update_variant_combo(DEFAULT_DEVICE)
        self._device_config_template.set_config_fields(device_config)
        self._old_device_name = device_config.get("name", "")

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(1600, 1000)

    def create_control_panel(self) -> QtWidgets.QGroupBox:
        self._control_box = QtWidgets.QGroupBox("Choose a Device Group")
        layout = QtWidgets.QGridLayout(self._control_box)

        group_label = QtWidgets.QLabel("Device Group:")
        layout.addWidget(group_label, 0, 0)

        group_combo = QtWidgets.QComboBox()
        group_combo.addItems(self._group_variants.keys())
        self._control_widgets["group_combo"] = group_combo
        layout.addWidget(group_combo, 1, 0)

        variant_label = QtWidgets.QLabel("Variants:")
        layout.addWidget(variant_label, 0, 1)

        variant_combo = QtWidgets.QComboBox()
        self._control_widgets["variant_combo"] = variant_combo
        layout.addWidget(variant_combo, 1, 1)

        group_combo.currentTextChanged.connect(self.update_variant_combo)
        variant_combo.currentTextChanged.connect(self.update_device_config_template)

        return self._control_box

    def update_variant_combo(self, group_name: str):
        variant_combo = self._control_widgets["variant_combo"]
        variant_combo.clear()
        variant_combo.addItems(self._group_variants.get(group_name, []))
        if variant_combo.count() <= 1:
            variant_combo.setEnabled(False)
        else:
            variant_combo.setEnabled(True)

    def update_device_config_template(self, variant_name: str):
        group_name = self._control_widgets["group_combo"].currentText()
        template_info = OPHYD_DEVICE_TEMPLATES.get(group_name, {}).get(variant_name, {})
        if template_info:
            self._device_config_template.change_template(template_info)
        else:
            self._device_config_template.change_template(
                OPHYD_DEVICE_TEMPLATES[DEFAULT_DEVICE][DEFAULT_DEVICE]
            )

    def _create_validation_dialog(self) -> QtWidgets.QProgressDialog:
        """
        Create and show a validation progress dialog while validating the device configuration.
        The dialog will be modal and prevent user interaction until validation is complete.
        """
        wait_dialog = QtWidgets.QProgressDialog(
            "Validating config... please wait", None, 0, 0, parent=self
        )
        wait_dialog.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        wait_dialog.setCancelButton(None)
        wait_dialog.setMinimumDuration(0)
        return wait_dialog

    def _create_and_run_ophyd_validation(self, config: dict[str, Any]) -> OphydValidation:
        """Run ophyd validation test on the current device configuration."""
        ophyd_validation = OphydValidation(parent=self)
        ophyd_validation.validation_completed.connect(self._handle_validation_result)
        ophyd_validation.multiple_validations_completed.connect(
            self._handle_devices_already_in_session_results
        )

        # NOTE Use singleShot here to ensure that the signal is emitted after all other scheduled
        # tasks in the event loop are processed. This avoids potential deadlocks. In particular,
        # this is relevant for the _wait_dialog exec which opens a modal dialog during validation
        # and therefore must not have the signal emitted immediately in the same event loop iteration.
        # Otherwise, the callback may be scheduled before the dialog is shown resulting in a deadlock.
        QtCore.QTimer.singleShot(
            0, lambda: ophyd_validation.change_device_configs([config], True, False)
        )
        return ophyd_validation

    @SafeSlot(list)
    def _handle_devices_already_in_session_results(
        self, validation_results: _ValidationResultIter
    ) -> None:
        """Handle completion if device is already in session."""
        if len(validation_results) != 1:
            logger.error(
                "Expected a single device validation result, but got multiple. Using first result."
            )
        result = validation_results[0] if len(validation_results) > 0 else None
        if result is None:
            logger.error(
                f"Received validation results: {validation_results} of unexpected length 0. Returning."
            )
            return
        device_config, config_status, connection_status, validation_msg = result
        self._handle_validation_result(
            device_config, config_status, connection_status, validation_msg
        )

    @SafeSlot(dict, int, int, str)
    def _handle_validation_result(
        self, device_config: dict, config_status: int, connection_status: int, validation_msg: str
    ):
        """Handle completion of validation."""
        try:
            if (
                DeviceModel.model_validate(device_config)
                == DeviceModel.model_validate(self._validation_result[0])
                and connection_status == ConnectionStatus.UNKNOWN.value
            ):
                # Config unchanged, we can reuse previous connection status. Only do this if the new
                # connection status is UNKNOWN as the current validation should not test the connection.
                connection_status = self._validation_result[2]
                validation_msg = self._validation_result[3]
        except Exception:
            logger.debug(
                f"Device config validation changed for config: {device_config} compared to previous validation. Using status from recent validation."
            )
        self._validation_result = (device_config, config_status, connection_status, validation_msg)
        if self._wait_dialog is not None:
            self._wait_dialog.accept()
            self._wait_dialog.close()
            self._wait_dialog.deleteLater()
            self._wait_dialog = None

    def _add_config(self):
        """
        Adding a config will always run a validation check of the config without a connection test.
        We will check if tests have already run, and reuse the information in case they also tested the connection to the device.
        """
        config = self._device_config_template.get_config_fields()

        # I. First we validate that the device name is valid, as this may create issues within the OphydValidation widget.
        # Validate device name first. If invalid, this should immediately block adding the device.
        if not validate_name(config.get("name", "")):
            msg_box = self._create_warning_message_box(
                "Invalid Device Name",
                f"Device is invalid, cannot be empty or contain spaces. Please provide a valid name. {config.get('name', '')!r}",
            )
            msg_box.exec()
            return

        # II. Next we will run the validation check of the config without connection test.
        # We will show a wait dialog while this is happening, and compare the results with the last known validation results.
        # If the config is unchanged, we will use the connection status results from the last validation.
        self._wait_dialog = self._create_validation_dialog()
        ophyd_validation: OphydValidation | None = None
        try:
            ophyd_validation = self._create_and_run_ophyd_validation(config)

            # NOTE If dialog was already closed, this means that a validation callback was already received
            # which closed the dialog. In this case, we skip exec to avoid deadlock. With the singleShot above,
            # this should not happen, but we keep the check for safety.
            if self._wait_dialog is not None:
                self._wait_dialog.exec()  # This will block until the validation is complete

            config, config_status, connection_status, validation_msg = self._validation_result

            if config_status == ConfigStatus.INVALID.value:
                msg_box = self._create_warning_message_box(
                    "Invalid Device Configuration",
                    f"Device configuration is invalid. Last known validation message:\n\nErrors:\n{self._validation_result[3]}",
                )
                msg_box.exec()
                return

            self.accepted_data.emit(
                config, config_status, connection_status, validation_msg, self._old_device_name
            )
            self.accept()
        finally:
            if ophyd_validation is not None:
                ophyd_validation.close()
                ophyd_validation.deleteLater()

    def _create_warning_message_box(self, title: str, text: str) -> QtWidgets.QMessageBox:
        msg_box = QtWidgets.QMessageBox(self)
        msg_box.setIcon(QtWidgets.QMessageBox.Warning)
        msg_box.setWindowTitle(title)
        msg_box.setText(text)
        return msg_box

    def _test_connection(self):
        config = self._device_config_template.get_config_fields()
        dialog = DeviceManagerOphydValidationDialog(self, config=config)
        result = dialog.exec()
        if result in (QtWidgets.QDialog.Accepted, QtWidgets.QDialog.Rejected):
            self.config_validation_result = dialog.validation_result

    def _reset_config(self):
        self._device_config_template.reset_to_defaults()

    def _reject_config(self):
        self.reject()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from bec_qthemes import apply_theme

    app = QtWidgets.QApplication(sys.argv)
    apply_theme("light")

    dialog = DeviceFormDialog()
    dialog.resize(1200, 800)
    dialog.show()
    sys.exit(app.exec())



import enum
import random
import time

from bec_qthemes import material_icon
from qtpy.QtCore import QSize, Qt, QTimer, Signal, Slot
from qtpy.QtGui import QBrush, QColor, QPainter, QPen
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.bec_widget import BECWidget

NUM_COLORS = {
    1: QColor("#f44336"),
    2: QColor("#9C27B0"),
    3: QColor("#3F51B5"),
    4: QColor("#03A9F4"),
    5: QColor("#00BCD4"),
    6: QColor("#4CAF50"),
    7: QColor("#E91E63"),
    8: QColor("#FF9800"),
}

LEVELS: dict[str, tuple[int, int]] = {"1": (8, 10), "2": (16, 40), "3": (24, 99)}


class GameStatus(enum.Enum):
    READY = 0
    PLAYING = 1
    FAILED = 2
    SUCCESS = 3


class Pos(QWidget):
    expandable = Signal(int, int)
    clicked = Signal()
    ohno = Signal()

    def __init__(self, x, y, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setFixedSize(QSize(20, 20))

        self.x = x
        self.y = y
        self.is_start = False
        self.is_mine = False
        self.adjacent_n = 0
        self.is_revealed = False
        self.is_flagged = False

    def reset(self):
        """Restore the tile to its original state before mine status is assigned"""
        self.is_start = False
        self.is_mine = False
        self.adjacent_n = 0

        self.is_revealed = False
        self.is_flagged = False

        self.update()

    def paintEvent(self, event):
        p = QPainter(self)

        r = event.rect()

        if self.is_revealed:
            color = self.palette().base().color()
            outer, inner = color, color
        else:
            outer, inner = (self.palette().highlightedText().color(), self.palette().text().color())

        p.fillRect(r, QBrush(inner))
        pen = QPen(outer)
        pen.setWidth(1)
        p.setPen(pen)
        p.drawRect(r)

        if self.is_revealed:
            if self.is_mine:
                p.drawPixmap(r, material_icon("experiment", convert_to_pixmap=True, filled=True))

            elif self.adjacent_n > 0:
                pen = QPen(NUM_COLORS[self.adjacent_n])
                p.setPen(pen)
                f = p.font()
                f.setBold(True)
                p.setFont(f)
                p.drawText(r, Qt.AlignHCenter | Qt.AlignVCenter, str(self.adjacent_n))

        elif self.is_flagged:
            p.drawPixmap(
                r,
                material_icon(
                    "flag",
                    size=(50, 50),
                    convert_to_pixmap=True,
                    filled=True,
                    color=self.palette().base().color(),
                ),
            )
        p.end()

    def flag(self):
        self.is_flagged = not self.is_flagged
        self.update()

        self.clicked.emit()

    def reveal(self):
        self.is_revealed = True
        self.update()

    def click(self):
        if not self.is_revealed:
            self.reveal()
            if self.adjacent_n == 0:
                self.expandable.emit(self.x, self.y)

        self.clicked.emit()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.MouseButton.RightButton and not self.is_revealed:
            self.flag()
            return

        if event.button() == Qt.MouseButton.LeftButton:
            self.click()
            if self.is_mine:
                self.ohno.emit()


class Minesweeper(BECWidget, QWidget):

    PLUGIN = True
    ICON_NAME = "videogame_asset"
    USER_ACCESS = []
    RPC = True

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        self._ui_initialised = False
        self._timer_start_num_seconds = 0
        self._set_level_params(LEVELS["1"])

        self._init_ui()
        self._init_map()

        self.update_status(GameStatus.READY)
        self.reset_map()
        self.update_status(GameStatus.READY)

    def _init_ui(self):
        if self._ui_initialised:
            return
        self._ui_initialised = True

        status_hb = QHBoxLayout()
        self.mines = QLabel()
        self.mines.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        f = self.mines.font()
        f.setPointSize(24)
        self.mines.setFont(f)

        self.reset_button = QPushButton()
        self.reset_button.setFixedSize(QSize(32, 32))
        self.reset_button.setIconSize(QSize(32, 32))
        self.reset_button.setFlat(True)
        self.reset_button.pressed.connect(self.reset_button_pressed)

        self.clock = QLabel()
        self.clock.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.clock.setFont(f)
        self._timer = QTimer()
        self._timer.timeout.connect(self.update_timer)
        self._timer.start(1000)  # 1 second timer
        self.mines.setText(f"{self.num_mines:03d}")
        self.clock.setText("000")

        status_hb.addWidget(self.mines)
        status_hb.addWidget(self.reset_button)
        status_hb.addWidget(self.clock)

        level_hb = QHBoxLayout()
        self.level_selector = QComboBox()
        self.level_selector.addItems(list(LEVELS.keys()))
        level_hb.addWidget(QLabel("Level: "))
        level_hb.addWidget(self.level_selector)
        self.level_selector.currentTextChanged.connect(self.change_level)

        vb = QVBoxLayout()
        vb.addLayout(level_hb)
        vb.addLayout(status_hb)

        self.grid = QGridLayout()
        self.grid.setSpacing(5)

        vb.addLayout(self.grid)
        self.setLayout(vb)

    def _init_map(self):
        """Redraw the grid of mines"""

        # Remove any previous grid items and reset the grid
        for i in reversed(range(self.grid.count())):
            w: Pos = self.grid.itemAt(i).widget()
            w.clicked.disconnect(self.on_click)
            w.expandable.disconnect(self.expand_reveal)
            w.ohno.disconnect(self.game_over)
            w.setParent(None)
            w.deleteLater()

        # Add positions to the map
        for x in range(0, self.b_size):
            for y in range(0, self.b_size):
                w = Pos(x, y)
                self.grid.addWidget(w, y, x)
                # Connect signal to handle expansion.
                w.clicked.connect(self.on_click)
                w.expandable.connect(self.expand_reveal)
                w.ohno.connect(self.game_over)

    def reset_map(self):
        """
        Reset the map and add new mines.
        """
        # Clear all mine positions
        for x in range(0, self.b_size):
            for y in range(0, self.b_size):
                w = self.grid.itemAtPosition(y, x).widget()
                w.reset()

        # Add mines to the positions
        positions = []
        while len(positions) < self.num_mines:
            x, y = (random.randint(0, self.b_size - 1), random.randint(0, self.b_size - 1))
            if (x, y) not in positions:
                w = self.grid.itemAtPosition(y, x).widget()
                w.is_mine = True
                positions.append((x, y))

        def get_adjacency_n(x, y):
            positions = self.get_surrounding(x, y)
            num_mines = sum(1 if w.is_mine else 0 for w in positions)

            return num_mines

        # Add adjacencies to the positions
        for x in range(0, self.b_size):
            for y in range(0, self.b_size):
                w = self.grid.itemAtPosition(y, x).widget()
                w.adjacent_n = get_adjacency_n(x, y)

        # Place starting marker
        while True:
            x, y = (random.randint(0, self.b_size - 1), random.randint(0, self.b_size - 1))
            w = self.grid.itemAtPosition(y, x).widget()
            # We don't want to start on a mine.
            if (x, y) not in positions:
                w = self.grid.itemAtPosition(y, x).widget()
                w.is_start = True

                # Reveal all positions around this, if they are not mines either.
                for w in self.get_surrounding(x, y):
                    if not w.is_mine:
                        w.click()
                break

    def get_surrounding(self, x, y):
        positions = []
        for xi in range(max(0, x - 1), min(x + 2, self.b_size)):
            for yi in range(max(0, y - 1), min(y + 2, self.b_size)):
                positions.append(self.grid.itemAtPosition(yi, xi).widget())
        return positions

    def get_num_hidden(self) -> int:
        """
        Get the number of hidden positions.
        """
        return sum(
            1
            for x in range(0, self.b_size)
            for y in range(0, self.b_size)
            if not self.grid.itemAtPosition(y, x).widget().is_revealed
        )

    def get_num_remaining_flags(self) -> int:
        """
        Get the number of remaining flags.
        """
        return self.num_mines - sum(
            1
            for x in range(0, self.b_size)
            for y in range(0, self.b_size)
            if self.grid.itemAtPosition(y, x).widget().is_flagged
        )

    def reset_button_pressed(self):
        match self.status:
            case GameStatus.PLAYING:
                self.game_over()
            case GameStatus.FAILED | GameStatus.SUCCESS:
                self.reset_map()

    def reveal_map(self):
        for x in range(0, self.b_size):
            for y in range(0, self.b_size):
                w = self.grid.itemAtPosition(y, x).widget()
                w.reveal()

    @Slot(str)
    def change_level(self, level: str):
        self._set_level_params(LEVELS[level])
        self._init_map()
        self.reset_map()

    @Slot(int, int)
    def expand_reveal(self, x, y):
        """
        Expand the reveal to the surrounding

        Args:
            x (int): The x position.
            y (int): The y position.
        """
        for xi in range(max(0, x - 1), min(x + 2, self.b_size)):
            for yi in range(max(0, y - 1), min(y + 2, self.b_size)):
                w = self.grid.itemAtPosition(yi, xi).widget()
                if not w.is_mine:
                    w.click()

    @Slot()
    def on_click(self):
        """
        Handle the click event. If the game is not started, start the game.
        """
        self.update_available_flags()
        if self.status != GameStatus.PLAYING:
            # First click.
            self.update_status(GameStatus.PLAYING)
            # Start timer.
            self._timer_start_num_seconds = int(time.time())
            return
        self.check_win()

    def update_available_flags(self):
        """
        Update the number of available flags.
        """
        self.mines.setText(f"{self.get_num_remaining_flags():03d}")

    def check_win(self):
        """
        Check if the game is won.
        """
        if self.get_num_hidden() == self.num_mines:
            self.update_status(GameStatus.SUCCESS)

    def update_status(self, status: GameStatus):
        """
        Update the status of the game.

        Args:
            status (GameStatus): The status of the game.
        """
        self.status = status
        match status:
            case GameStatus.READY:
                icon = material_icon(icon_name="add", convert_to_pixmap=False)
            case GameStatus.PLAYING:
                icon = material_icon(icon_name="smart_toy", convert_to_pixmap=False)
            case GameStatus.FAILED:
                icon = material_icon(icon_name="error", convert_to_pixmap=False)
            case GameStatus.SUCCESS:
                icon = material_icon(icon_name="celebration", convert_to_pixmap=False)
        self.reset_button.setIcon(icon)

    def update_timer(self):
        """
        Update the timer.
        """
        if self.status == GameStatus.PLAYING:
            num_seconds = int(time.time()) - self._timer_start_num_seconds
            self.clock.setText(f"{num_seconds:03d}")

    def game_over(self):
        """Cause the game to end early"""
        self.reveal_map()
        self.update_status(GameStatus.FAILED)

    def _set_level_params(self, level: tuple[int, int]):
        self.b_size, self.num_mines = level

    def cleanup(self):
        self._timer.stop()
        super().cleanup()


if __name__ == "__main__":
    from bec_widgets.utils.colors import apply_theme

    app = QApplication([])
    apply_theme("light")
    widget = Minesweeper()
    widget.show()

    app.exec_()



"""Module for custom input widgets used in device configuration templates."""

from ast import literal_eval
from typing import Any, Callable

from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from pydantic import BaseModel, ConfigDict
from qtpy import QtWidgets

from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.widgets.control.scan_control.scan_group_box import ScanDoubleSpinBox
from bec_widgets.widgets.utility.toggle.toggle import ToggleSwitch

logger = bec_logger.logger


def _try_literal_eval(value: str) -> Any:
    """Consolidated function for literal evaluation of a value."""
    if value in ["true", "True"]:
        return True
    if value in ["false", "False"]:
        return False
    if value == "":
        return ""
    try:
        return literal_eval(f"{value}")
    except ValueError:
        return value
    except Exception:
        logger.warning(f"Could not literal_eval value: {value}, returning as string")
        return value


class InputLineEdit(QtWidgets.QLineEdit):
    """
    Custom QLineEdit for input fields with validation.

    Args:
        parent (QtWidgets.QWidget, optional): Parent widget. Defaults to None.
        config_field (str, optional): Configuration field name. Defaults to "no_field_specified"
        required (bool, optional): Whether the field is required. Defaults to True.
        placeholder_text (str, optional): Placeholder text for the input field. Defaults to "".
    """

    def __init__(
        self,
        parent=None,
        config_field: str = "no_field_specified",
        required: bool = True,
        placeholder_text: str = "",
    ):
        super().__init__(parent)
        self._config_field = config_field
        self._colors = get_accent_colors()
        self._required = required
        self.textChanged.connect(self._update_input_field_style)
        self._validation_callbacks: list[Callable[[bool], str]] = []
        self.setPlaceholderText(placeholder_text)
        self._update_input_field_style()

    def register_validation_callback(self, callback: Callable[[str], bool]) -> None:
        """
        Register a custom validation callback.

        Args:
            callback (Callable[[str], bool]): A function that takes the input string
                and returns True if valid, False otherwise.
        """
        self._validation_callbacks.append(callback)

    def apply_theme(self, theme: str) -> None:
        """Apply the theme to the widget."""
        self._colors = get_accent_colors()
        self._update_input_field_style()

    def _update_input_field_style(self) -> None:
        """Update the input field style based on validation."""
        name = self.text()
        if not self.is_valid_input(name) and self._required is True:
            self.setStyleSheet(f"border: 1px solid {self._colors.emergency.name()};")
            return
        self.setStyleSheet("")
        return

    def is_valid_input(self, name: str) -> bool:
        """Validate the input string using plugin helper."""
        name = name.strip()  # Remove leading/trailing whitespace
        # Run registered validation callbacks
        for callback in self._validation_callbacks:
            try:
                valid = callback(name)
            except Exception as exc:
                logger.warning(
                    f"Validation callback raised an exception: {exc}. Defaulting to valid"
                )
                valid = True
            if not valid:
                return False
        if not self._required:
            return True
        if not name:
            return False
        return True


class OnFailureComboBox(QtWidgets.QComboBox):
    """Custom QComboBox for the onFailure input field."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.addItems(["buffer", "retry", "raise"])


class ReadoutPriorityComboBox(QtWidgets.QComboBox):
    """Custom QComboBox for the readoutPriority input field."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.addItems(["monitored", "baseline", "async", "continuous", "on_request"])


class LimitInputWidget(QtWidgets.QWidget):
    """Custom widget for inputting limits as a tuple (min, max)."""

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent)
        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(4)

        # Colors
        self._colors = get_accent_colors()

        self.min_input = ScanDoubleSpinBox(self, arg_name="min_limit", default=0.0)
        self.min_input.setPrefix("Min: ")
        self.min_input.setEnabled(False)
        self.min_input.setRange(-1e12, 1e12)
        self._layout.addWidget(self.min_input)

        self.max_input = ScanDoubleSpinBox(self, arg_name="max_limit", default=0.0)
        self.max_input.setPrefix("Max: ")
        self.max_input.setRange(-1e12, 1e12)
        self.max_input.setEnabled(False)
        self._layout.addWidget(self.max_input)

        # Add validity checks
        self.min_input.valueChanged.connect(self._check_valid_inputs)
        self.max_input.valueChanged.connect(self._check_valid_inputs)

        # Add checkbox to enable/disable limits
        self.enable_toggle = ToggleSwitch(self)
        self.enable_toggle.setToolTip("Enable editing limits")
        self.enable_toggle.setChecked(False)
        self.enable_toggle.enabled.connect(self._toggle_limits_enabled)
        self._layout.addWidget(self.enable_toggle)

    def reset_defaults(self) -> None:
        """Reset limits to default values."""
        self.min_input.setValue(0.0)
        self.max_input.setValue(0.0)
        self.enable_toggle.setChecked(False)

    def _is_valid_limit(self) -> bool:
        """Check if the current limits are valid (min < max)."""
        return self.min_input.value() <= self.max_input.value()

    def _check_valid_inputs(self) -> None:
        """Check if the current inputs are valid and update styles accordingly."""
        if not self._is_valid_limit():
            self.min_input.setStyleSheet(f"border: 1px solid {self._colors.emergency.name()};")
            self.max_input.setStyleSheet(f"border: 1px solid {self._colors.emergency.name()};")
        else:
            self.min_input.setStyleSheet("")
            self.max_input.setStyleSheet("")

    def _toggle_limits_enabled(self, enable: bool) -> None:
        """Enable or disable the limit inputs based on the checkbox state."""
        self.min_input.setEnabled(enable)
        self.max_input.setEnabled(enable)

    def get_limits(self) -> list[float, float]:
        """Return the limits as a list [min, max]."""
        min_val = self.min_input.value()
        max_val = self.max_input.value()
        return [min_val, max_val]

    def set_limits(self, limits: tuple) -> None:
        """Set the limits from a tuple (min, max)."""
        checked_state = self.enable_toggle.isChecked()
        if not checked_state:
            self.enable_toggle.setChecked(True)
        self.min_input.setValue(limits[0])
        self.max_input.setValue(limits[1])
        self.enable_toggle.setChecked(checked_state)


class ParameterValueWidget(QtWidgets.QWidget):
    """Custom QTreeWidget for user parameters input field."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(4)
        self.tree_widget = QtWidgets.QTreeWidget(self)
        self._layout.addWidget(self.tree_widget)
        self.tree_widget.setColumnCount(2)
        self.tree_widget.setHeaderLabels(["Parameter", "Value"])
        self.tree_widget.setIndentation(0)
        self.tree_widget.setRootIsDecorated(False)
        header = self.tree_widget.header()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        self._add_tool_buttons()

    def clear_widget(self) -> None:
        """Clear all tags."""
        for i in reversed(range(self.tree_widget.topLevelItemCount())):
            item = self.tree_widget.topLevelItem(i)
            index = self.tree_widget.indexOfTopLevelItem(item)
            if index != -1:
                self.tree_widget.takeTopLevelItem(index)

    def _add_tool_buttons(self) -> None:
        """Add tool buttons for adding/removing parameter lines."""
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.setSpacing(4)
        self._layout.addLayout(button_layout)
        self._button_add = QtWidgets.QPushButton(self)
        self._button_add.setIcon(material_icon("add", size=(16, 16), convert_to_pixmap=False))
        self._button_add.setToolTip("Add parameter")
        self._button_add.clicked.connect(self._add_button_clicked)
        button_layout.addWidget(self._button_add)

        self._button_remove = QtWidgets.QPushButton(self)
        self._button_remove.setIcon(material_icon("remove", size=(16, 16), convert_to_pixmap=False))
        self._button_remove.setToolTip("Remove selected parameter")
        self._button_remove.clicked.connect(self.remove_parameter_line)
        button_layout.addWidget(self._button_remove)

    def _add_button_clicked(self, *args, **kwargs) -> None:
        """Handle the add button click event."""
        self.add_parameter_line()

    def add_parameter_line(self, parameter: str | None = None, value: str | None = None) -> None:
        """Add a new row with editable Parameter/Value QLineEdits."""
        item = QtWidgets.QTreeWidgetItem(self.tree_widget)
        self.tree_widget.addTopLevelItem(item)

        # Parameter field
        param_edit = QtWidgets.QLineEdit(self.tree_widget)
        param_edit.setPlaceholderText("Parameter")
        self.tree_widget.setItemWidget(item, 0, param_edit)

        # Value field
        value_edit = QtWidgets.QLineEdit(self.tree_widget)
        value_edit.setPlaceholderText("Value")
        self.tree_widget.setItemWidget(item, 1, value_edit)
        if parameter is not None:
            param_edit.setText(str(parameter))
        if value is not None:
            value_edit.setText(str(value))

    def remove_parameter_line(self) -> None:
        """Remove the selected row."""
        selected_items = self.tree_widget.selectedItems()
        for item in selected_items:
            index = self.tree_widget.indexOfTopLevelItem(item)
            if index != -1:
                self.tree_widget.takeTopLevelItem(index)

    # ---------------------------------------------------------------------

    def parameters(self) -> dict:
        """Return all parameters as a dictionary {parameter: value}."""
        result = {}
        for i in range(self.tree_widget.topLevelItemCount()):
            item = self.tree_widget.topLevelItem(i)
            param_edit = self.tree_widget.itemWidget(item, 0)
            value_edit = self.tree_widget.itemWidget(item, 1)
            if param_edit and value_edit:
                key = param_edit.text().strip()
                val = value_edit.text().strip()
                if key and val:
                    result[key] = _try_literal_eval(val)
        return result


class DeviceTagsWidget(QtWidgets.QWidget):
    """Custom QTreeWidget for deviceTags input field."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(4)
        self.tree_widget = QtWidgets.QTreeWidget(self)
        self._layout.addWidget(self.tree_widget)
        self.tree_widget.setColumnCount(1)
        self.tree_widget.setHeaderLabels(["Tags"])
        self.tree_widget.setIndentation(0)
        self.tree_widget.setRootIsDecorated(False)
        self._add_tool_buttons()

    def clear_widget(self) -> None:
        """Clear all tags."""
        for i in reversed(range(self.tree_widget.topLevelItemCount())):
            item = self.tree_widget.topLevelItem(i)
            index = self.tree_widget.indexOfTopLevelItem(item)
            if index != -1:
                self.tree_widget.takeTopLevelItem(index)

    def _add_tool_buttons(self) -> None:
        """Add tool buttons for adding/removing parameter lines."""
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.setSpacing(4)
        self._layout.addLayout(button_layout)
        self._button_add = QtWidgets.QPushButton(self)
        self._button_add.setIcon(material_icon("add", size=(16, 16), convert_to_pixmap=False))
        self._button_add.setToolTip("Add parameter")
        self._button_add.clicked.connect(self._add_button_clicked)
        button_layout.addWidget(self._button_add)

        self._button_remove = QtWidgets.QPushButton(self)
        self._button_remove.setIcon(material_icon("remove", size=(16, 16), convert_to_pixmap=False))
        self._button_remove.setToolTip("Remove selected parameter")
        self._button_remove.clicked.connect(self.remove_parameter_line)
        button_layout.addWidget(self._button_remove)

    def _add_button_clicked(self, *args, **kwargs) -> None:
        """Handle the add button click event."""
        self.add_parameter_line()

    def add_parameter_line(self, parameter: str | None = None) -> None:
        """Add a new row with editable Tag QLineEdit."""
        item = QtWidgets.QTreeWidgetItem(self.tree_widget)
        self.tree_widget.addTopLevelItem(item)

        # Tag field
        param_edit = QtWidgets.QLineEdit(self.tree_widget)
        param_edit.setPlaceholderText("Tag")
        self.tree_widget.setItemWidget(item, 0, param_edit)
        if parameter is not None:
            param_edit.setText(str(parameter))

    def remove_parameter_line(self) -> None:
        """Remove the selected row."""
        selected_items = self.tree_widget.selectedItems()
        for item in selected_items:
            index = self.tree_widget.indexOfTopLevelItem(item)
            if index != -1:
                self.tree_widget.takeTopLevelItem(index)

    # ---------------------------------------------------------------------

    def parameters(self) -> list[str]:
        """Return all parameters as a list of tags."""
        result = []
        for i in range(self.tree_widget.topLevelItemCount()):
            item = self.tree_widget.topLevelItem(i)
            param_edit = self.tree_widget.itemWidget(item, 0)
            if param_edit:
                tag = param_edit.text().strip()
                if tag:
                    result.append(tag)
        return result


# Validation callback for name field
def validate_name(name: str) -> bool:
    """Check that the name does not contain spaces."""
    if " " in name:
        return False
    if not name.replace("_", "").isalnum():
        return False
    return True


# Validation callback for deviceClass field
def validate_device_cls(name: str) -> bool:
    """Check that the name does not contain spaces."""
    if " " in name:
        return False
    if not name.replace("_", "").replace(".", "").isalnum():
        return False
    return True


def validate_prefix(value: str) -> bool:
    """Check that the prefix does not contain spaces."""
    if " " in value:
        return False
    if not value.replace("_", "").replace(".", "").replace("-", "").replace(":", "").isalnum():
        return False
    return True


class DeviceConfigField(BaseModel):
    """Pydantic model for device configuration fields."""

    label: str
    widget_cls: type[QtWidgets.QWidget]
    required: bool = False
    static: bool = False
    placeholder_text: str | None = None
    validation_callback: list[Callable[[str], bool]] | None = None
    default: Any = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


DEVICE_FIELDS = {
    "name": DeviceConfigField(
        label="Name",
        widget_cls=InputLineEdit,
        required=True,
        placeholder_text="Device name (no spaces or special characters)",
        validation_callback=[validate_name],
    ),
    "deviceClass": DeviceConfigField(
        label="Device Class",
        widget_cls=InputLineEdit,
        required=True,
        placeholder_text="Device class (no spaces or special characters)",
        validation_callback=[validate_device_cls],
    ),
    "description": DeviceConfigField(
        label="Description",
        widget_cls=QtWidgets.QTextEdit,
        required=False,
        placeholder_text="Short device description",
    ),
    "enabled": DeviceConfigField(
        label="Enabled", widget_cls=ToggleSwitch, required=False, default=True
    ),
    "readOnly": DeviceConfigField(
        label="Read Only", widget_cls=ToggleSwitch, required=False, default=False
    ),
    "softwareTrigger": DeviceConfigField(
        label="Software Trigger", widget_cls=ToggleSwitch, required=False, default=False
    ),
    "readoutPriority": DeviceConfigField(
        label="Readout Priority", widget_cls=ReadoutPriorityComboBox, default="baseline"
    ),
    "onFailure": DeviceConfigField(
        label="On Failure", widget_cls=OnFailureComboBox, default="retry"
    ),
    "userParameter": DeviceConfigField(
        label="User Parameters", widget_cls=ParameterValueWidget, static=False
    ),
    "deviceTags": DeviceConfigField(label="Device Tags", widget_cls=DeviceTagsWidget, static=False),
}

DEVICE_CONFIG_FIELDS = {
    "prefix": DeviceConfigField(
        label="Prefix",
        widget_cls=InputLineEdit,
        static=False,
        placeholder_text="EPICS IOC prefix, e.g. X25DA-ES1-MOT:",
        validation_callback=[validate_prefix],
    ),
    "read_pv": DeviceConfigField(
        label="Read PV",
        widget_cls=InputLineEdit,
        static=False,
        placeholder_text="EPICS read PV: e.g. X25DA-ES1-MOT:GET",
        validation_callback=[validate_prefix],
    ),
    "write_pv": DeviceConfigField(
        label="Write PV",
        widget_cls=InputLineEdit,
        static=False,
        placeholder_text="EPICS write PV (if different from read_pv): e.g. X25DA-ES1-MOT:SET",
        validation_callback=[validate_prefix],
    ),
    "limits": DeviceConfigField(label="Limits", widget_cls=LimitInputWidget, static=False),
    "DEFAULT": DeviceConfigField(label="DEFAULT FIELD", widget_cls=InputLineEdit, static=False),
}



from collections import defaultdict
from types import NoneType, SimpleNamespace
from typing import Optional

from bec_lib.endpoints import MessageEndpoints
from pydantic import BaseModel, Field
from qtpy.QtCore import Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils import ConnectionConfig
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import apply_theme, get_accent_colors
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.widgets.control.buttons.stop_button.stop_button import StopButton
from bec_widgets.widgets.control.scan_control.scan_group_box import ScanGroupBox
from bec_widgets.widgets.editors.scan_metadata.scan_metadata import ScanMetadata
from bec_widgets.widgets.utility.toggle.toggle import ToggleSwitch


class ScanParameterConfig(BaseModel):
    name: str
    args: Optional[list] = Field(None)
    kwargs: Optional[dict] = Field(None)


class ScanControlConfig(ConnectionConfig):
    default_scan: Optional[str] = Field(None)
    allowed_scans: Optional[list] = Field(None)
    scans: Optional[dict[str, ScanParameterConfig]] = defaultdict(dict)


class ScanControl(BECWidget, QWidget):
    """
    Widget to submit new scans to the queue.
    """

    USER_ACCESS = ["attach", "detach", "screenshot"]
    PLUGIN = True
    ICON_NAME = "tune"
    ARG_BOX_POSITION: int = 2

    scan_started = Signal()
    scan_selected = Signal(str)
    device_selected = Signal(str)
    scan_args = Signal(list)

    def __init__(
        self,
        parent=None,
        client=None,
        config: ScanControlConfig | dict | None = None,
        gui_id: str | None = None,
        allowed_scans: list | None = None,
        default_scan: str | None = None,
        **kwargs,
    ):
        if config is None:
            config = ScanControlConfig(
                widget_class=self.__class__.__name__, allowed_scans=allowed_scans
            )
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self._hide_add_remove_buttons = False

        # Client from BEC + shortcuts to device manager and scans
        self.get_bec_shortcuts()

        # Main layout
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(5, 5, 5, 5)
        self.arg_box = None
        self.kwarg_boxes = []
        self.expert_mode = False  # TODO implement in the future versions
        self.previous_scan = None
        self.last_scan_found = None

        # Widget Default Parameters
        self.config.default_scan = default_scan
        self.config.allowed_scans = allowed_scans

        self._scan_metadata: dict | None = None
        self._metadata_form = ScanMetadata(parent=self)
        self._hide_arg_box = False
        self._hide_kwarg_boxes = False
        self._hide_scan_control_buttons = False
        self._hide_metadata = False
        self._hide_scan_selection_combobox = False

        # Create and set main layout
        self._init_UI()

    def _init_UI(self):
        """
        Initializes the UI of the scan control widget. Create the top box for scan selection and populate scans to main combobox.
        """
        palette = get_accent_colors()
        if palette is None:
            palette = SimpleNamespace(
                default=QColor("blue"),
                success=QColor("green"),
                warning=QColor("orange"),
                emergency=QColor("red"),
            )
        # Scan selection box
        self.scan_selection_group = QWidget(self)
        QVBoxLayout(self.scan_selection_group)
        scan_selection_layout = QHBoxLayout()
        self.comboBox_scan_selection_label = QLabel("Scan:", self.scan_selection_group)
        self.comboBox_scan_selection = QComboBox(self.scan_selection_group)
        scan_selection_layout.addWidget(self.comboBox_scan_selection_label, 0)
        scan_selection_layout.addWidget(self.comboBox_scan_selection, 1)
        self.scan_selection_group.layout().addLayout(scan_selection_layout)

        # Label to reload the last scan parameters within scan selection group box
        self.toggle_layout = QHBoxLayout()
        self.toggle_layout.addSpacerItem(
            QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        )
        self.last_scan_label = QLabel("Restore last scan parameters", self.scan_selection_group)
        self.toggle = ToggleSwitch(parent=self.scan_selection_group, checked=False)
        self.toggle.enabled.connect(self.request_last_executed_scan_parameters)
        self.toggle_layout.addWidget(self.last_scan_label)
        self.toggle_layout.addWidget(self.toggle)
        self.scan_selection_group.layout().addLayout(self.toggle_layout)
        self.scan_selection_group.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed
        )
        self.layout.addWidget(self.scan_selection_group)

        # Scan control (Run/Stop) buttons
        self.scan_control_group = QWidget(self)
        self.scan_control_group.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed
        )
        self.button_layout = QHBoxLayout(self.scan_control_group)
        self.button_run_scan = QPushButton("Start", self.scan_control_group)
        self.button_run_scan.setProperty("variant", "success")
        self.button_stop_scan = StopButton(parent=self.scan_control_group)
        self.button_layout.addWidget(self.button_run_scan)
        self.button_layout.addWidget(self.button_stop_scan)
        self.layout.addWidget(self.scan_control_group)

        # Default scan from config
        if self.config.default_scan is not None:
            self.comboBox_scan_selection.setCurrentText(self.config.default_scan)

        # Connect signals
        self.comboBox_scan_selection.view().pressed.connect(self.save_current_scan_parameters)
        self.comboBox_scan_selection.currentIndexChanged.connect(self.on_scan_selection_changed)
        self.button_run_scan.clicked.connect(self.run_scan)

        self.scan_selected.connect(self.scan_select)

        # Initialize scan selection
        self.populate_scans()

        # Append metadata form
        self._add_metadata_form()

        self.layout.addStretch()

    def _add_metadata_form(self):
        # Wrap metadata form in a group box
        self._metadata_group = QGroupBox("Scan Metadata", self)
        self._metadata_group.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        metadata_layout = QVBoxLayout(self._metadata_group)
        metadata_layout.addWidget(self._metadata_form)

        self.layout.addWidget(self._metadata_group)
        self._metadata_form.update_with_new_scan(self.comboBox_scan_selection.currentText())
        self.scan_selected.connect(self._metadata_form.update_with_new_scan)
        self._metadata_form.form_data_updated.connect(self.update_scan_metadata)
        self._metadata_form.form_data_cleared.connect(self.update_scan_metadata)
        self._metadata_form.validate_form()

    def populate_scans(self):
        """Populates the scan selection combo box with available scans from BEC session."""
        self.available_scans = self.client.connector.get(
            MessageEndpoints.available_scans()
        ).resource
        if self.config.allowed_scans is None:
            supported_scans = ["ScanBase", "SyncFlyScanBase", "AsyncFlyScanBase"]
            allowed_scans = [
                scan_name
                for scan_name, scan_info in self.available_scans.items()
                if scan_info["base_class"] in supported_scans and len(scan_info["gui_config"]) > 0
            ]

        else:
            allowed_scans = self.config.allowed_scans
        self.comboBox_scan_selection.addItems(allowed_scans)

    def on_scan_selection_changed(self, index: int):
        """Callback for scan selection combo box"""
        selected_scan_name = self.comboBox_scan_selection.currentText()
        self.scan_selected.emit(selected_scan_name)
        self.request_last_executed_scan_parameters()
        self.restore_scan_parameters(selected_scan_name)

    @SafeSlot()
    @SafeSlot(bool)
    def request_last_executed_scan_parameters(self, *_):
        """
        Requests the last executed scan parameters from BEC and restores them to the scan control widget.
        """
        self.last_scan_found = False
        if not self.toggle.checked:
            return

        current_scan = self.comboBox_scan_selection.currentText()
        history = (
            self.client.connector.xread(
                MessageEndpoints.scan_history(), from_start=True, user_id=self.object_name
            )
            or []
        )

        for scan in reversed(history):
            scan_data = scan.get("data")
            if not scan_data:
                continue

            if scan_data.scan_name != current_scan:
                continue

            ri = getattr(scan_data, "request_inputs", {}) or {}
            args_list = ri.get("arg_bundle", [])
            if args_list and self.arg_box:
                self.arg_box.set_parameters(args_list)

            inputs = ri.get("inputs", {})
            kwargs = ri.get("kwargs", {})
            merged = {**inputs, **kwargs}
            if merged and self.kwarg_boxes:
                for box in self.kwarg_boxes:
                    box.set_parameters(merged)

            self.last_scan_found = True
            break

    @SafeProperty(str)
    def current_scan(self):
        """Returns the scan name for the currently selected scan."""
        return self.comboBox_scan_selection.currentText()

    @current_scan.setter
    def current_scan(self, scan_name: str):
        """Sets the current scan to the given scan name.

        Args:
            scan_name(str): Name of the scan to set as current.
        """
        if scan_name not in self.available_scans:
            return
        self.comboBox_scan_selection.setCurrentText(scan_name)

    @SafeSlot(str)
    def set_current_scan(self, scan_name: str):
        """Slot for setting the current scan to the given scan name.

        Args:
            scan_name(str): Name of the scan to set as current.
        """
        self.current_scan = scan_name

    @SafeProperty(bool)
    def hide_arg_box(self):
        """Property to hide the argument box."""
        return self._hide_arg_box

    @hide_arg_box.setter
    def hide_arg_box(self, hide: bool):
        """Setter for the hide_arg_box property.

        Args:
            hide(bool): Hide or show the argument box.
        """
        self._hide_arg_box = hide
        if self.arg_box is not None:
            self.arg_box.setVisible(not hide)

    @SafeProperty(bool)
    def hide_kwarg_boxes(self):
        """Property to hide the keyword argument boxes."""
        return self._hide_kwarg_boxes

    @hide_kwarg_boxes.setter
    def hide_kwarg_boxes(self, hide: bool):
        """Setter for the hide_kwarg_boxes property.

        Args:
            hide(bool): Hide or show the keyword argument boxes.
        """
        self._hide_kwarg_boxes = hide
        if len(self.kwarg_boxes) > 0:
            for box in self.kwarg_boxes:
                box.setVisible(not hide)

    @SafeProperty(bool)
    def hide_scan_control_buttons(self):
        """Property to hide the scan control buttons."""
        return self._hide_scan_control_buttons

    @hide_scan_control_buttons.setter
    def hide_scan_control_buttons(self, hide: bool):
        """Setter for the hide_scan_control_buttons property.

        Args:
            hide(bool): Hide or show the scan control buttons.
        """
        self._hide_scan_control_buttons = hide
        self.show_scan_control_buttons(not hide)

    @SafeProperty(bool)
    def hide_metadata(self):
        """Property to hide the metadata form."""
        return self._hide_metadata

    @hide_metadata.setter
    def hide_metadata(self, hide: bool):
        """Setter for the hide_metadata property.

        Args:
            hide(bool): Hide or show the metadata form.
        """
        self._hide_metadata = hide
        self._metadata_form.setVisible(not hide)

    @SafeProperty(bool)
    def hide_optional_metadata(self):
        """Property to hide the optional metadata form."""
        return self._metadata_form.hide_optional_metadata

    @hide_optional_metadata.setter
    def hide_optional_metadata(self, hide: bool):
        """Setter for the hide_optional_metadata property.

        Args:
            hide(bool): Hide or show the optional metadata form.
        """
        self._metadata_form.hide_optional_metadata = hide

    @SafeSlot(bool)
    def show_scan_control_buttons(self, show: bool):
        """Shows or hides the scan control buttons."""
        self._hide_scan_control_buttons = not show
        self.scan_control_group.setVisible(show)

    @SafeProperty(bool)
    def hide_scan_selection_combobox(self):
        """Property to hide the scan selection combobox."""
        return self._hide_scan_selection_combobox

    @hide_scan_selection_combobox.setter
    def hide_scan_selection_combobox(self, hide: bool):
        """Setter for the hide_scan_selection_combobox property.

        Args:
            hide(bool): Hide or show the scan selection combobox.
        """
        self._hide_scan_selection_combobox = hide
        self.show_scan_selection_combobox(not hide)

    @SafeSlot(bool)
    def show_scan_selection_combobox(self, show: bool):
        """Shows or hides the scan selection combobox."""
        self._hide_scan_selection_combobox = not show
        self.scan_selection_group.setVisible(show)

    @SafeSlot(str)
    def scan_select(self, scan_name: str):
        """
        Slot for scan selection. Updates the scan control layout based on the selected scan.

        Args:
            scan_name(str): Name of the selected scan.
        """
        self.reset_layout()
        selected_scan_info = self.available_scans.get(scan_name, {})

        gui_config = selected_scan_info.get("gui_config", {})
        self.arg_group = gui_config.get("arg_group", None)
        self.kwarg_groups = gui_config.get("kwarg_groups", None)

        if bool(self.arg_group["arg_inputs"]):
            self.add_arg_group(self.arg_group)
        if len(self.kwarg_groups) > 0:
            self.add_kwargs_boxes(self.kwarg_groups)

        self.update()
        self.adjustSize()

    @SafeProperty(bool)
    def hide_add_remove_buttons(self):
        """Property to hide the add_remove buttons."""
        return self._hide_add_remove_buttons

    @hide_add_remove_buttons.setter
    def hide_add_remove_buttons(self, hide: bool):
        """Setter for the hide_add_remove_buttons property.

        Args:
            hide(bool): Hide or show the add_remove buttons.
        """
        self._hide_add_remove_buttons = hide
        if self.arg_box is not None:
            self.arg_box.hide_add_remove_buttons = hide

    def add_kwargs_boxes(self, groups: list):
        """
        Adds the given gui_groups to the scan control layout.

        Args:
            groups(list): List of dictionaries containing the gui_group information.
        """
        position = self.ARG_BOX_POSITION + (1 if self.arg_box is not None else 0)
        for group in groups:
            box = ScanGroupBox(box_type="kwargs", config=group)
            box.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
            self.layout.insertWidget(position + len(self.kwarg_boxes), box)
            self.kwarg_boxes.append(box)
            box.setVisible(not self._hide_kwarg_boxes)

    def add_arg_group(self, group: dict):
        """
        Adds the given gui_groups to the scan control layout.

        Args:
        """
        self.arg_box = ScanGroupBox(box_type="args", config=group)
        self.arg_box.device_selected.connect(self.emit_device_selected)
        self.arg_box.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        self.arg_box.hide_add_remove_buttons = self._hide_add_remove_buttons
        self.layout.insertWidget(self.ARG_BOX_POSITION, self.arg_box)
        self.arg_box.setVisible(not self._hide_arg_box)

    @SafeSlot(str)
    def emit_device_selected(self, dev_names):
        """
        Emit the signal to inform about selected device(s)

        "dev_names" is a string separated by space, in case of multiple devices
        """
        self._selected_devices = dev_names
        self.device_selected.emit(dev_names)

    def reset_layout(self):
        """Clears the scan control layout from GuiGroups and ArgGroups boxes."""
        if self.arg_box is not None:
            self.layout.removeWidget(self.arg_box)
            self.arg_box.deleteLater()
            self.arg_box = None
        if self.kwarg_boxes != []:
            self.remove_kwarg_boxes()

    def remove_kwarg_boxes(self):
        for box in self.kwarg_boxes:
            self.layout.removeWidget(box)
            box.deleteLater()
        self.kwarg_boxes = []

    def get_scan_parameters(self, bec_object: bool = True):
        """
        Returns the scan parameters for the selected scan.

        Args:
            bec_object(bool): If True, returns the BEC object for the scan parameters such as device objects.
        """
        args = []
        kwargs = {}
        if self.arg_box is not None:
            args = self.arg_box.get_parameters(bec_object)
        for box in self.kwarg_boxes:
            box_kwargs = box.get_parameters(bec_object)
            kwargs.update(box_kwargs)
        if self._scan_metadata is not None:
            kwargs["metadata"] = self._scan_metadata
        return args, kwargs

    def restore_scan_parameters(self, scan_name: str):
        """
        Restores the scan parameters for the given scan name

        Args:
            scan_name(str): Name of the scan to restore the parameters for.
        """
        if self.last_scan_found is True:
            return
        scan_params = self.config.scans.get(scan_name, None)
        if scan_params is None and self.previous_scan is None:
            return

        if scan_params is None and self.previous_scan is not None:
            previous_scan_params = self.config.scans.get(self.previous_scan, None)
            self._restore_kwargs(previous_scan_params.kwargs)
            return

        if scan_params.args is not None and self.arg_box is not None:
            self.arg_box.set_parameters(scan_params.args)

        self._restore_kwargs(scan_params.kwargs)

    def _restore_kwargs(self, scan_kwargs: dict):
        """Restores the kwargs for the given scan parameters."""
        if scan_kwargs is not None and self.kwarg_boxes is not None:
            for box in self.kwarg_boxes:
                box.set_parameters(scan_kwargs)

    def save_current_scan_parameters(self):
        """Saves the current scan parameters to the scan control config for further use."""
        scan_name = self.comboBox_scan_selection.currentText()
        self.previous_scan = scan_name
        args, kwargs = self.get_scan_parameters(False)
        scan_params = ScanParameterConfig(name=scan_name, args=args, kwargs=kwargs)
        self.config.scans[scan_name] = scan_params

    @SafeSlot(dict)
    @SafeSlot(NoneType)
    def update_scan_metadata(self, md: dict | None):
        self._scan_metadata = md
        if md is None:
            self.button_run_scan.setEnabled(False)
        else:
            self.button_run_scan.setEnabled(True)

    @SafeSlot(popup_error=True)
    def run_scan(self):
        """Starts the selected scan with the given parameters."""
        args, kwargs = self.get_scan_parameters()
        self.scan_args.emit(args)
        scan_function = getattr(self.scans, self.comboBox_scan_selection.currentText())
        if callable(scan_function):
            self.scan_started.emit()
            scan_function(*args, **kwargs)

    def cleanup(self):
        """Cleanup the scan control widget."""
        self.button_stop_scan.cleanup()
        if self.arg_box:
            for widget in self.arg_box.widgets:
                if hasattr(widget, "cleanup"):
                    widget.cleanup()
        for kwarg_box in self.kwarg_boxes:
            for widget in kwarg_box.widgets:
                if hasattr(widget, "cleanup"):
                    widget.cleanup()
        super().cleanup()


# Application example
if __name__ == "__main__":  # pragma: no cover
    app = QApplication([])
    scan_control = ScanControl()

    apply_theme("dark")
    window = scan_control
    window.show()
    app.exec()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import shiboken6
from bec_lib.logger import bec_logger
from qtpy.QtCore import QBuffer, QByteArray, QIODevice, QObject, Qt
from qtpy.QtGui import QFont, QPixmap
from qtpy.QtWidgets import QApplication, QFileDialog, QLabel, QVBoxLayout, QWidget

import bec_widgets.widgets.containers.qt_ads as QtAds
from bec_widgets.cli.rpc.rpc_register import RPCRegister
from bec_widgets.utils.bec_connector import BECConnector, ConnectionConfig
from bec_widgets.utils.busy_loader import install_busy_loader
from bec_widgets.utils.error_popups import SafeConnect, SafeSlot
from bec_widgets.utils.rpc_decorator import rpc_timeout
from bec_widgets.utils.widget_io import WidgetHierarchy
from bec_widgets.widgets.utility.spinner.spinner import SpinnerWidget

if TYPE_CHECKING:  # pragma: no cover
    from bec_widgets.utils.busy_loader import BusyLoaderOverlay
    from bec_widgets.widgets.containers.dock import BECDock

logger = bec_logger.logger


class BECWidget(BECConnector):
    """Mixin class for all BEC widgets, to handle cleanup"""

    # The icon name is the name of the icon in the icon theme, typically a name taken
    # from fonts.google.com/icons. Override this in subclasses to set the icon name.
    ICON_NAME = "widgets"
    USER_ACCESS = ["remove", "attach", "detach"]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        client=None,
        config: ConnectionConfig = None,
        gui_id: str | None = None,
        theme_update: bool = False,
        start_busy: bool = False,
        **kwargs,
    ):
        """
        Base class for all BEC widgets. This class should be used as a mixin class for all BEC widgets, e.g.:


        >>> class MyWidget(BECWidget, QWidget):
        >>>     def __init__(self, parent=None, client=None, config=None, gui_id=None):
        >>>         super().__init__(parent=parent, client=client, config=config, gui_id=gui_id)


        Args:
            client(BECClient, optional): The BEC client.
            config(ConnectionConfig, optional): The connection configuration.
            gui_id(str, optional): The GUI ID.
            theme_update(bool, optional): Whether to subscribe to theme updates. Defaults to False. When set to True, the
                widget's apply_theme method will be called when the theme changes.
        """
        super().__init__(client=client, config=config, gui_id=gui_id, **kwargs)
        if not isinstance(self, QObject):
            raise RuntimeError(f"{repr(self)} is not a subclass of QWidget")
        if theme_update:
            logger.debug(f"Subscribing to theme updates for {self.__class__.__name__}")
            self._connect_to_theme_change()

        # Initialize optional busy loader overlay utility (lazy by default)
        self._busy_overlay: "BusyLoaderOverlay" | None = None
        self._busy_state_widget: QWidget | None = None

        self._loading = False
        self._busy_overlay = self._install_busy_loader()
        if start_busy and isinstance(self, QWidget):
            self._show_busy_overlay()
            self._loading = True

    def _connect_to_theme_change(self):
        """Connect to the theme change signal."""
        qapp = QApplication.instance()
        if hasattr(qapp, "theme"):
            SafeConnect(self, qapp.theme.theme_changed, self._update_theme)

    @SafeSlot(str)
    @SafeSlot()
    def _update_theme(self, theme: str | None = None):
        """Update the theme."""
        if theme is None:
            qapp = QApplication.instance()
            if hasattr(qapp, "theme"):
                theme = qapp.theme.theme
            else:
                theme = "dark"
        self._update_overlay_theme(theme)
        self.apply_theme(theme)

    def create_busy_state_widget(self) -> QWidget:
        """
        Method to create a custom busy state widget to be shown in the busy overlay.
        Child classes should overrid this method to provide a custom widget if desired.

        Returns:
            QWidget: The custom busy state widget.

        NOTE:
            The implementation here is a SpinnerWidget with a "Loading..." label. This is the default
            busy state widget for all BECWidgets. However, child classes with specific needs for the
            busy state can easily overrite this method to provide a custom widget. The signature of
            the method must be preserved to ensure compatibility with the busy overlay system. If
            the widget provides a 'cleanup' method, it will be called when the overlay is cleaned up.

            The widget may connect to the _busy_overlay signals foreground_color_changed and
            scrim_color_changed to update its colors when the theme changes.
        """

        # Widget
        class BusyStateWidget(QWidget):
            def __init__(self, parent=None):
                super().__init__(parent)
                # label
                label = QLabel("Loading...", self)
                label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
                f = QFont(label.font())
                f.setBold(True)
                f.setPointSize(f.pointSize() + 1)
                label.setFont(f)

                # spinner
                spinner = SpinnerWidget(self)
                spinner.setFixedSize(42, 42)

                # Layout
                lay = QVBoxLayout(self)
                lay.setContentsMargins(24, 24, 24, 24)
                lay.setSpacing(10)
                lay.addStretch(1)
                lay.addWidget(spinner, 0, Qt.AlignHCenter)
                lay.addWidget(label, 0, Qt.AlignHCenter)
                lay.addStretch(1)
                self.setLayout(lay)

            def showEvent(self, event):
                """Show event to start the spinner."""
                super().showEvent(event)
                for child in self.findChildren(SpinnerWidget):
                    child.start()

            def hideEvent(self, event):
                """Hide event to stop the spinner."""
                super().hideEvent(event)
                for child in self.findChildren(SpinnerWidget):
                    child.stop()

        widget = BusyStateWidget(self)
        return widget

    def _install_busy_loader(self) -> "BusyLoaderOverlay" | None:
        """
        Create the busy overlay on demand and cache it in _busy_overlay.
        Returns the overlay instance or None if not a QWidget.
        """
        if not isinstance(self, QWidget):
            return None
        overlay = getattr(self, "_busy_overlay", None)
        if overlay is None:

            overlay = install_busy_loader(target=self, start_loading=False)
            self._busy_overlay = overlay

        # Create and set the busy state widget
        self._busy_state_widget = self.create_busy_state_widget()
        self._busy_overlay.set_widget(self._busy_state_widget)
        return overlay

    def _show_busy_overlay(self) -> None:
        """Create and attach the loading overlay to this widget if QWidget is present."""
        if not isinstance(self, QWidget):
            return
        if self._busy_overlay is not None:
            self._busy_overlay.setGeometry(self.rect())  # pylint: disable=no-member
            self._busy_overlay.raise_()
            self._busy_overlay.show()

    def set_busy(self, enabled: bool) -> None:
        """
        Set the busy state of the widget. This will show or hide the loading overlay, which will
        block user interaction with the widget and show the busy_state_widget if provided. Per
        default, the busy state widget is a spinner with "Loading..." text.

        Args:
            enabled(bool): Whether to enable the busy state.
        """
        if not isinstance(self, QWidget):
            return
        # If not yet installed, install the busy overlay now together with the busy state widget
        if self._busy_overlay is None:
            self._busy_overlay = self._install_busy_loader()
        if enabled:
            self._show_busy_overlay()
        else:
            self._busy_overlay.hide()
        self._loading = bool(enabled)

    def is_busy(self) -> bool:
        """
        Check if the loading overlay is enabled.

        Returns:
            bool: True if the loading overlay is enabled, False otherwise.
        """
        return bool(getattr(self, "_loading", False))

    @SafeSlot(str)
    def apply_theme(self, theme: str):
        """
        Apply the theme to the widget.

        Args:
            theme(str, optional): The theme to be applied.
        """

    def _update_overlay_theme(self, theme: str):
        try:
            overlay = getattr(self, "_busy_overlay", None)
            if overlay is not None:
                overlay._update_palette()
        except Exception:
            logger.warning(f"Failed to apply theme {theme} to {self}")

    def get_help_md(self) -> str:
        """
        Method to override in subclasses to provide help text in markdown format.

        Returns:
            str: The help text in markdown format.
        """
        return ""

    @SafeSlot()
    @SafeSlot(str)
    @rpc_timeout(None)
    def screenshot(self, file_name: str | None = None):
        """
        Take a screenshot of the dock area and save it to a file.
        """
        if not isinstance(self, QWidget):
            logger.error("Cannot take screenshot of non-QWidget instance")
            return

        screenshot = self.grab()
        if file_name is None:
            file_name, _ = QFileDialog.getSaveFileName(
                self,
                "Save Screenshot",
                f"bec_{datetime.now().strftime('%Y%m%d_%H%M%S')}.png",
                "PNG Files (*.png);;JPEG Files (*.jpg *.jpeg);;All Files (*)",
            )
        if not file_name:
            return
        screenshot.save(file_name)
        logger.info(f"Screenshot saved to {file_name}")

    def screenshot_bytes(
        self,
        *,
        max_width: int | None = None,
        max_height: int | None = None,
        fmt: str = "PNG",
        quality: int = -1,
    ) -> QByteArray:
        """
        Grab this widget, optionally scale to a max size, and return encoded image bytes.

        If max_width/max_height are omitted (the default), capture at full resolution.

        Args:
            max_width(int, optional): Maximum width of the screenshot.
            max_height(int, optional): Maximum height of the screenshot.
            fmt(str, optional): Image format (e.g., "PNG", "JPEG").
            quality(int, optional): Image quality (0-100), -1 for default.

        Returns:
            QByteArray: The screenshot image bytes.
        """
        if not isinstance(self, QWidget):
            return QByteArray()

        if not hasattr(self, "grab"):
            raise RuntimeError(f"Cannot take screenshot of non-QWidget instance: {repr(self)}")

        pixmap: QPixmap = self.grab()
        if pixmap.isNull():
            return QByteArray()
        if max_width is not None or max_height is not None:
            w = max_width if max_width is not None else pixmap.width()
            h = max_height if max_height is not None else pixmap.height()
            pixmap = pixmap.scaled(
                w, h, Qt.AspectRatioMode.KeepAspectRatio, Qt.QSmoothTransformation
            )
        ba = QByteArray()
        buf = QBuffer(ba)
        buf.open(QIODevice.OpenModeFlag.WriteOnly)
        pixmap.save(buf, fmt, quality)
        buf.close()
        return ba

    def attach(self):
        dock = WidgetHierarchy.find_ancestor(self, QtAds.CDockWidget)
        if dock is None:
            return

        if not dock.isFloating():
            return
        dock.dockManager().addDockWidget(QtAds.DockWidgetArea.RightDockWidgetArea, dock)

    def detach(self):
        """
        Detach the widget from its parent dock widget (if widget is in the dock), making it a floating widget.
        """
        dock = WidgetHierarchy.find_ancestor(self, QtAds.CDockWidget)
        if dock is None:
            return
        if dock.isFloating():
            return
        dock.setFloating()

    def cleanup(self):
        """Cleanup the widget."""
        with RPCRegister.delayed_broadcast():
            # All widgets need to call super().cleanup() in their cleanup method
            logger.info(f"Registry cleanup for widget {self.__class__.__name__}")
            self.rpc_register.remove_rpc(self)
        children = self.findChildren(BECWidget)
        for child in children:
            if not shiboken6.isValid(child):
                # If the child is not valid, it means it has already been deleted
                continue
            child.close()
            child.deleteLater()

        # Tear down busy overlay explicitly to stop spinner and remove filters
        overlay = getattr(self, "_busy_overlay", None)
        if overlay is not None and shiboken6.isValid(overlay):
            try:
                overlay.hide()
                filt = getattr(overlay, "_filter", None)
                if filt is not None and shiboken6.isValid(filt):
                    try:
                        self.removeEventFilter(filt)
                    except Exception as exc:
                        logger.warning(f"Failed to remove event filter from busy overlay: {exc}")

                # Cleanup the overlay widget. This will call cleanup on the custom widget if present.

                overlay.cleanup()
                overlay.deleteLater()
            except Exception as exc:
                logger.warning(f"Failed to delete busy overlay: {exc}")

    def closeEvent(self, event):
        """Wrap the close even to ensure the rpc_register is cleaned up."""
        try:
            if not self._destroyed:
                self.cleanup()
                self._destroyed = True
        finally:
            super().closeEvent(event)  # pylint: disable=no-member



from __future__ import annotations

from typing import Literal

from bec_lib import bec_logger

from bec_widgets.widgets.containers.auto_update.auto_updates import AutoUpdates
from bec_widgets.widgets.containers.dock_area.dock_area import BECDockArea

logger = bec_logger.logger


def dock_area(
    object_name: str | None = None, startup_profile: str | Literal["restore", "skip"] | None = None
) -> BECDockArea:
    """
    Create an advanced dock area using Qt Advanced Docking System.

    Args:
        object_name(str): The name of the advanced dock area.
        startup_profile(str | Literal["restore", "skip"] | None): Startup mode for
            the workspace:
              - None: start empty
              - "restore": restore last used profile
              - "skip": do not initialize profile state
              - "<name>": load specific profile

    Returns:
        BECDockArea: The created advanced dock area.

    """

    widget = BECDockArea(
        object_name=object_name,
        root_widget=True,
        profile_namespace="bec",
        startup_profile=startup_profile,
    )
    logger.info(f"Created advanced dock area with startup_profile: {startup_profile}")
    return widget


def auto_update_dock_area(object_name: str | None = None) -> AutoUpdates:
    """
    Create a dock area with auto update enabled.

    Args:
        object_name(str): The name of the dock area.

    Returns:
        BECDockArea: The created dock area.
    """
    _auto_update = AutoUpdates(object_name=object_name)
    return _auto_update



from bec_lib.codecs import BECCodec
from bec_lib.serialization import msgpack
from qtpy.QtCore import QPointF


class QPointFEncoder(BECCodec):
    obj_type = QPointF

    @staticmethod
    def encode(obj: QPointF) -> list[float]:
        """Encode a QPointF object to a list of floats."""
        return [obj.x(), obj.y()]

    @staticmethod
    def decode(type_name: str, data: list[float]) -> list[float]:
        """No-op function since QPointF is encoded as a list of floats."""
        return data


def register_serializer_extension():
    """
    Register the serializer extension for the BECConnector.
    """
    if not msgpack.is_registered(QPointF):
        msgpack.register(QPointF, QPointFEncoder.encode, QPointFEncoder.decode)



from __future__ import annotations

from typing import Callable

from qtpy.QtCore import Qt
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QComboBox, QSizePolicy

from bec_widgets import SafeSlot
from bec_widgets.utils.toolbars.actions import MaterialIconAction, WidgetAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle, ToolbarComponents
from bec_widgets.utils.toolbars.connections import BundleConnection
from bec_widgets.widgets.containers.dock_area.profile_utils import list_quick_profiles


class ProfileComboBox(QComboBox):
    """Custom combobox that displays icons for read-only profiles."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self._quick_provider: Callable[[], list[str]] = list_quick_profiles

    def set_quick_profile_provider(self, provider: Callable[[], list[str]]) -> None:
        self._quick_provider = provider

    def refresh_profiles(
        self, active_profile: str | None = None, show_empty_profile: bool = False
    ) -> None:
        """
        Refresh the profile list and ensure the active profile is visible.

        Args:
            active_profile(str | None): The currently active profile name.
            show_empty_profile(bool): If True, show an explicit empty unsaved workspace entry.
        """

        current_text = active_profile or self.currentText()
        self.blockSignals(True)
        self.clear()

        quick_profiles = self._quick_provider()
        quick_set = set(quick_profiles)

        items: list[str] = []
        if show_empty_profile:
            items.append("")

        if active_profile and active_profile not in quick_set:
            items.append(active_profile)

        for profile in quick_profiles:
            if profile not in items:
                items.append(profile)

        if active_profile and active_profile not in quick_set:
            # keep active profile at the top when not in quick list
            items.remove(active_profile)
            insert_pos = 1 if show_empty_profile else 0
            items.insert(insert_pos, active_profile)

        for profile in items:
            self.addItem(profile)
            idx = self.count() - 1

            # Reset any custom styling
            self.setItemData(idx, None, Qt.ItemDataRole.FontRole)
            self.setItemData(idx, None, Qt.ItemDataRole.ToolTipRole)
            self.setItemData(idx, None, Qt.ItemDataRole.ForegroundRole)

            if profile == "":
                self.setItemData(idx, "Unsaved empty workspace", Qt.ItemDataRole.ToolTipRole)
                if active_profile is None:
                    font = QFont(self.font())
                    font.setItalic(True)
                    self.setItemData(idx, font, Qt.ItemDataRole.FontRole)
                    self.setCurrentIndex(idx)
                continue

            if active_profile and profile == active_profile:
                tooltip = "Active workspace profile"
                if profile not in quick_set:
                    font = QFont(self.font())
                    font.setItalic(True)
                    font.setBold(True)
                    self.setItemData(idx, font, Qt.ItemDataRole.FontRole)
                    self.setItemData(
                        idx, self.palette().highlight().color(), Qt.ItemDataRole.ForegroundRole
                    )
                    tooltip = "Active profile (not in quick select)"
                self.setItemData(idx, tooltip, Qt.ItemDataRole.ToolTipRole)
                self.setCurrentIndex(idx)
            elif profile not in quick_set:
                self.setItemData(idx, "Not in quick select", Qt.ItemDataRole.ToolTipRole)

        # Restore selection if possible
        if show_empty_profile and active_profile is None:
            empty_idx = self.findText("")
            if empty_idx >= 0:
                self.setCurrentIndex(empty_idx)
        else:
            index = self.findText(current_text)
            if index >= 0:
                self.setCurrentIndex(index)

        self.blockSignals(False)
        if active_profile and self.currentText() != active_profile:
            idx = self.findText(active_profile)
            if idx >= 0:
                self.setCurrentIndex(idx)
        if show_empty_profile and self.currentText() == "":
            self.setToolTip("Unsaved empty workspace")
        elif active_profile and active_profile not in quick_set:
            self.setToolTip("Active profile is not in quick select")
        else:
            self.setToolTip("")


def workspace_bundle(components: ToolbarComponents, enable_tools: bool = True) -> ToolbarBundle:
    """
    Creates a workspace toolbar bundle for BECDockArea.

    Args:
        components (ToolbarComponents): The components to be added to the bundle.

    Returns:
        ToolbarBundle: The workspace toolbar bundle.
    """
    # Workspace combo
    combo = ProfileComboBox(parent=components.toolbar)
    combo.setVisible(enable_tools)
    components.add_safe("workspace_combo", WidgetAction(widget=combo, adjust_size=False))

    components.add_safe(
        "save_workspace",
        MaterialIconAction(
            icon_name="save",
            tooltip="Save Current Workspace",
            checkable=False,
            parent=components.toolbar,
        ),
    )
    components.get_action("save_workspace").action.setVisible(enable_tools)

    components.add_safe(
        "reset_default_workspace",
        MaterialIconAction(
            icon_name="undo",
            tooltip="Refresh Current Workspace",
            checkable=False,
            parent=components.toolbar,
        ),
    )
    components.get_action("reset_default_workspace").action.setVisible(enable_tools)

    components.add_safe(
        "manage_workspaces",
        MaterialIconAction(
            icon_name="manage_accounts", tooltip="Manage", checkable=True, parent=components.toolbar
        ),
    )
    components.get_action("manage_workspaces").action.setVisible(enable_tools)

    bundle = ToolbarBundle("workspace", components)
    bundle.add_action("workspace_combo")
    bundle.add_action("save_workspace")
    bundle.add_action("reset_default_workspace")
    bundle.add_action("manage_workspaces")
    return bundle


class WorkspaceConnection(BundleConnection):
    """
    Connection class for workspace actions in BECDockArea.
    """

    def __init__(self, components: ToolbarComponents, target_widget=None):
        super().__init__(parent=components.toolbar)
        self.bundle_name = "workspace"
        self.components = components
        self.target_widget = target_widget
        if not hasattr(self.target_widget, "workspace_is_locked"):
            raise AttributeError("Target widget must implement 'workspace_is_locked'.")
        self._connected = False

    def connect(self):
        self._connected = True
        # Connect the action to the target widget's method
        save_action = self.components.get_action("save_workspace").action
        if save_action.isVisible():
            save_action.triggered.connect(self.target_widget.save_profile_dialog)

        self.components.get_action("workspace_combo").widget.currentTextChanged.connect(
            self.target_widget.load_profile
        )

        reset_action = self.components.get_action("reset_default_workspace").action
        if reset_action.isVisible():
            reset_action.triggered.connect(self._reset_workspace_to_default)

        manage_action = self.components.get_action("manage_workspaces").action
        if manage_action.isVisible():
            manage_action.triggered.connect(self.target_widget.show_workspace_manager)

    def disconnect(self):
        if not self._connected:
            return
        # Disconnect the action from the target widget's method
        save_action = self.components.get_action("save_workspace").action
        if save_action.isVisible():
            save_action.triggered.disconnect(self.target_widget.save_profile_dialog)
        self.components.get_action("workspace_combo").widget.currentTextChanged.disconnect(
            self.target_widget.load_profile
        )

        reset_action = self.components.get_action("reset_default_workspace").action
        if reset_action.isVisible():
            reset_action.triggered.disconnect(self._reset_workspace_to_default)

        manage_action = self.components.get_action("manage_workspaces").action
        if manage_action.isVisible():
            manage_action.triggered.disconnect(self.target_widget.show_workspace_manager)
        self._connected = False

    @SafeSlot()
    def _reset_workspace_to_default(self):
        """
        Refreshes the current workspace.
        """
        self.target_widget.restore_user_profile_from_default()



from typing import Final

# Denotes a MIME type for JSON-encoded list of device config dictionaries
MIME_DEVICE_CONFIG: Final[str] = "application/x-bec_device_config"

# Custom user roles
SORT_KEY_ROLE: Final[int] = 117
CONFIG_DATA_ROLE: Final[int] = 118

# TODO 882 keep in sync with headers in device_table_view.py
HEADERS_HELP_MD: dict[str, str] = {
    "valid": {
        "long": "\n".join(
            [
                "## Valid",
                "The current configuration status of the device. Can be one of the following values: ",
                "### **VALID** \n The device configuration is valid and can be used.",
                "### **INVALID** \n The device configuration is invalid.",
                "### **UNKNOWN** \n The device configuration has not been validated yet.",
            ]
        ),
        "short": "Validation status of the device configuration.",
    },
    "connect": {
        "long": "\n".join(
            [
                "## Connect",
                "The current connection status of the device. Can be one of the following values: ",
                "### **CONNECTED** \n The device is connected and in current session.",
                "### **CAN_CONNECT** \n The connection to the device has been validated. It's not yet loaded in the current session.",
                "### **CANNOT_CONNECT** \n The connection to the device could not be established.",
                "### **UNKNOWN** \n The connection status of the device is unknown.",
            ]
        ),
        "short": "Connection status of the device.",
    },
    "name": {
        "long": "\n".join(["## Name ", "The name of the device."]),
        "short": "Name of the device.",
    },
    "deviceClass": {
        "long": "\n".join(
            [
                "## Device Class",
                "The device class specifies the type of the device. It will be used to create the instance.",
            ]
        ),
        "short": "Python class for the device.",
    },
    "readoutPriority": {
        "long": "\n".join(
            [
                "## Readout Priority",
                "The readout priority of the device. Can be one of the following values: ",
                "### **monitored** \n The monitored priority is used for devices that are read out during the scan (i.e. at every step) and whose value may change during the scan.",
                "### **baseline** \n The baseline priority is used for devices that are read out at the beginning of the scan and whose value does not change during the scan.",
                "### **async** \n The async priority is used for devices that are asynchronous to the monitored devices, and send their data independently.",
                "### **continuous** \n The continuous priority is used for devices that are read out continuously during the scan.",
                "### **on_request** \n The on_request priority is used for devices that should not be read out during the scan, yet are configured to be read out manually.",
            ]
        ),
        "short": "Readout priority of the device for scans in BEC.",
    },
    "deviceTags": {
        "long": "\n".join(
            [
                "## Device Tags",
                "A list of tags associated with the device. Tags can be used to group devices and filter them in the device manager.",
            ]
        ),
        "short": "Tags associated with the device.",
    },
    "enabled": {
        "long": "\n".join(
            [
                "## Enabled",
                "Indicator whether the device is enabled or disabled. Disabled devices can not be used.",
            ]
        ),
        "short": "Enabled status of the device.",
    },
    "readOnly": {
        "long": "\n".join(
            ["## Read Only", "Indicator that a device is read-only or can be modified."]
        ),
        "short": "Read-only status of the device.",
    },
    "onFailure": {
        "long": "\n".join(
            [
                "## On Failure",
                "Specifies the behavior of the device in case of a failure. Can be one of the following values: ",
                "### **buffer** \n The device readback will fall back to the last known value.",
                "### **retry** \n The device readback will be retried once, and raises an error if it fails again.",
                "### **raise** \n The device readback will raise immediately.",
            ]
        ),
        "short": "On failure behavior of the device.",
    },
    "softwareTrigger": {
        "long": "\n".join(
            [
                "## Software Trigger",
                "Indicator whether the device receives a software trigger from BEC during a scan.",
            ]
        ),
        "short": "Software trigger status of the device.",
    },
    "description": {
        "long": "\n".join(["## Description", "A short description of the device."]),
        "short": "Description of the device.",
    },
}



from __future__ import annotations

import os

from qtpy.QtWidgets import QFrame, QScrollArea, QVBoxLayout

from bec_widgets.utils import UILoader
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.settings_dialog import SettingWidget


class HeatmapSettings(SettingWidget):
    def __init__(self, parent=None, target_widget=None, popup=False, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        # This is a settings widget that depends on the target widget
        # and should mirror what is in the target widget.
        # Saving settings for this widget could result in recursively setting the target widget.
        self.setProperty("skip_settings", True)

        current_path = os.path.dirname(__file__)
        if popup:
            form = UILoader().load_ui(
                os.path.join(current_path, "heatmap_settings_horizontal.ui"), self
            )
        else:
            form = UILoader().load_ui(
                os.path.join(current_path, "heatmap_settings_vertical.ui"), self
            )

        self.target_widget = target_widget
        self.popup = popup

        # # Scroll area
        self.scroll_area = QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setWidget(form)

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.scroll_area)
        self.ui = form

        self.fetch_all_properties()

        self.target_widget.heatmap_property_changed.connect(self.fetch_all_properties)
        if popup is False:
            self.ui.button_apply.clicked.connect(self.accept_changes)

        self.ui.device_x.setFocus()

    @SafeSlot()
    def fetch_all_properties(self):
        """
        Fetch all properties from the target widget and update the settings widget.
        """
        if not self.target_widget:
            return

        # Get properties from the target widget
        color_map = getattr(self.target_widget, "color_map", None)

        # Default values for device properties
        device_x, signal_x = None, None
        device_y, signal_y = None, None
        device_z, signal_z = None, None

        # Safely access device properties
        if hasattr(self.target_widget, "_image_config") and self.target_widget._image_config:
            config = self.target_widget._image_config

            if hasattr(config, "device_x") and config.device_x:
                device_x = getattr(config.device_x, "device", None)
                signal_x = getattr(config.device_x, "signal", None)

            if hasattr(config, "device_y") and config.device_y:
                device_y = getattr(config.device_y, "device", None)
                signal_y = getattr(config.device_y, "signal", None)

            if hasattr(config, "device_z") and config.device_z:
                device_z = getattr(config.device_z, "device", None)
                signal_z = getattr(config.device_z, "signal", None)

        # Apply the properties to the settings widget
        if hasattr(self.ui, "color_map"):
            self.ui.color_map.colormap = color_map

        if hasattr(self.ui, "device_x"):
            self.ui.device_x.set_device(device_x)
        if hasattr(self.ui, "signal_x") and signal_x is not None:
            self.ui.signal_x.set_to_obj_name(signal_x)

        if hasattr(self.ui, "device_y"):
            self.ui.device_y.set_device(device_y)
        if hasattr(self.ui, "signal_y") and signal_y is not None:
            self.ui.signal_y.set_to_obj_name(signal_y)

        if hasattr(self.ui, "device_z"):
            self.ui.device_z.set_device(device_z)
        if hasattr(self.ui, "signal_z") and signal_z is not None:
            self.ui.signal_z.set_to_obj_name(signal_z)

        if hasattr(self.ui, "interpolation"):
            self.ui.interpolation.setCurrentText(
                getattr(self.target_widget._image_config, "interpolation", "linear")
            )
        if hasattr(self.ui, "oversampling_factor"):
            self.ui.oversampling_factor.setValue(
                getattr(self.target_widget._image_config, "oversampling_factor", 1.0)
            )
        if hasattr(self.ui, "enforce_interpolation"):
            self.ui.enforce_interpolation.setChecked(
                getattr(self.target_widget._image_config, "enforce_interpolation", False)
            )

    @SafeSlot()
    def accept_changes(self):
        """
        Apply all properties from the settings widget to the target widget.
        """
        device_x = self.ui.device_x.currentText()
        signal_x = self.ui.signal_x.get_signal_name()
        device_y = self.ui.device_y.currentText()
        signal_y = self.ui.signal_y.get_signal_name()
        device_z = self.ui.device_z.currentText()
        signal_z = self.ui.signal_z.get_signal_name()
        validate_bec = self.ui.validate_bec.checked
        color_map = self.ui.color_map.colormap
        interpolation = self.ui.interpolation.currentText()
        oversampling_factor = self.ui.oversampling_factor.value()
        enforce_interpolation = self.ui.enforce_interpolation.isChecked()

        self.target_widget.plot(
            device_x=device_x,
            device_y=device_y,
            device_z=device_z,
            signal_x=signal_x,
            signal_y=signal_y,
            signal_z=signal_z,
            color_map=color_map,
            validate_bec=validate_bec,
            interpolation=interpolation,
            oversampling_factor=oversampling_factor,
            enforce_interpolation=enforce_interpolation,
            reload=True,
        )

    def cleanup(self):
        self.ui.device_x.close()
        self.ui.device_x.deleteLater()
        self.ui.signal_x.close()
        self.ui.signal_x.deleteLater()
        self.ui.device_y.close()
        self.ui.device_y.deleteLater()
        self.ui.signal_y.close()
        self.ui.signal_y.deleteLater()
        self.ui.device_z.close()
        self.ui.device_z.deleteLater()
        self.ui.signal_z.close()
        self.ui.signal_z.deleteLater()
        self.ui.interpolation.close()
        self.ui.interpolation.deleteLater()



import time
from types import SimpleNamespace

from bec_qthemes import material_icon
from qtpy.QtCore import Property, Qt, Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QDialog,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.colors import get_accent_colors


class LedLabel(QLabel):
    success_led = "color: white;border-radius: 10;background-color: qlineargradient(spread:pad, x1:0.145, y1:0.16, x2:1, y2:1, stop:0 %s, stop:1 %s);"
    emergency_led = "color: white;border-radius: 10;background-color: qlineargradient(spread:pad, x1:0.145, y1:0.16, x2:0.92, y2:0.988636, stop:0 %s, stop:1 %s);"
    warning_led = "color: white;border-radius: 10;background-color: qlineargradient(spread:pad, x1:0.232, y1:0.272, x2:0.98, y2:0.959773, stop:0 %s, stop:1 %s);"
    default_led = "color: white;border-radius: 10;background-color: qlineargradient(spread:pad, x1:0.04, y1:0.0565909, x2:0.799, y2:0.795, stop:0 %s, stop:1 %s);"

    def __init__(self, parent=None):
        super().__init__(parent)

        self.palette = get_accent_colors()
        if self.palette is None:
            # no theme!
            self.palette = SimpleNamespace(
                default=QColor("blue"),
                success=QColor("green"),
                warning=QColor("orange"),
                emergency=QColor("red"),
            )
        self.setState("default")
        self.setFixedSize(20, 20)

    def setState(self, state: str):
        match state:
            case "success":
                r, g, b, a = self.palette.success.getRgb()
                self.setStyleSheet(
                    LedLabel.success_led
                    % (
                        f"rgba({r},{g},{b},{a})",
                        f"rgba({int(r*0.8)},{int(g*0.8)},{int(b*0.8)},{a})",
                    )
                )
            case "default":
                r, g, b, a = self.palette.default.getRgb()
                self.setStyleSheet(
                    LedLabel.default_led
                    % (
                        f"rgba({r},{g},{b},{a})",
                        f"rgba({int(r*0.8)},{int(g*0.8)},{int(b*0.8)},{a})",
                    )
                )
            case "warning":
                r, g, b, a = self.palette.warning.getRgb()
                self.setStyleSheet(
                    LedLabel.warning_led
                    % (
                        f"rgba({r},{g},{b},{a})",
                        f"rgba({int(r*0.8)},{int(g*0.8)},{int(b*0.8)},{a})",
                    )
                )
            case "emergency":
                r, g, b, a = self.palette.emergency.getRgb()
                self.setStyleSheet(
                    LedLabel.emergency_led
                    % (
                        f"rgba({r},{g},{b},{a})",
                        f"rgba({int(r*0.8)},{int(g*0.8)},{int(b*0.8)},{a})",
                    )
                )
            case unknown_state:
                raise ValueError(
                    f"Unknown state {repr(unknown_state)}, must be one of default, success, warning or emergency"
                )


class PopupDialog(QDialog):
    def __init__(self, content_widget):
        self.parent = content_widget.parent()
        self.content_widget = content_widget

        super().__init__(self.parent)

        self.setAttribute(Qt.WA_DeleteOnClose)

        self.content_widget.setParent(self)
        QVBoxLayout(self)
        self.layout().addWidget(self.content_widget)
        self.content_widget.setVisible(True)

    def closeEvent(self, event):
        self.content_widget.setVisible(False)
        self.content_widget.setParent(self.parent)
        self.done(True)


class CompactPopupWidget(QWidget):
    """Container widget, that can display its content or have a compact form,
    in this case clicking on a small button pops the contained widget up.

    In the compact form, a LED-like indicator shows a status indicator.
    """

    expand = Signal(bool)

    def __init__(self, parent=None, layout=QVBoxLayout):
        super().__init__(parent)

        self._popup_window = None
        self._expand_popup = True

        QVBoxLayout(self)
        self.compact_view_widget = QWidget(self)
        self.compact_view_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        QHBoxLayout(self.compact_view_widget)
        self.compact_view_widget.layout().setSpacing(5)
        self.compact_view_widget.layout().setContentsMargins(0, 0, 0, 0)
        self.compact_view_widget.layout().addSpacerItem(
            QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Fixed)
        )
        self.compact_label = QLabel(self.compact_view_widget)
        self.compact_status = LedLabel(self.compact_view_widget)
        self.compact_show_popup = QToolButton(self.compact_view_widget)
        self.compact_show_popup.setIcon(
            material_icon(icon_name="expand_content", size=(10, 10), convert_to_pixmap=False)
        )
        self.compact_view_widget.layout().addWidget(self.compact_label)
        self.compact_view_widget.layout().addWidget(self.compact_status)
        self.compact_view_widget.layout().addWidget(self.compact_show_popup)
        self.compact_view_widget.setVisible(False)
        self.layout().addWidget(self.compact_view_widget)
        self.container = QWidget(self)
        self.layout().addWidget(self.container)
        self.container.setVisible(True)
        layout(self.container)
        self.layout = self.container.layout()
        self._compact_view = False

        self.compact_show_popup.clicked.connect(self.show_popup)

    def set_global_state(self, state: str):
        """Set the LED-indicator state

        The LED indicator represents the 'global' state. State can be one of the
        following: "default", "success", "warning", "emergency"
        """
        self.compact_status.setState(state)

    def show_popup(self):
        """Display the contained widgets in a popup dialog"""
        if self._expand_popup:
            # show popup
            self._popup_window = PopupDialog(self.container)
            self._popup_window.show()
            self._popup_window.finished.connect(lambda: self.expand.emit(False))
            self.expand.emit(True)
        else:
            if self.compact_view:
                # expand in place
                self.compact_view = False
                self.compact_view_widget.setVisible(True)
                self.compact_label.setVisible(False)
                self.compact_status.setVisible(False)
                self.compact_show_popup.setIcon(
                    material_icon(
                        icon_name="collapse_content", size=(10, 10), convert_to_pixmap=False
                    )
                )
                self.expand.emit(True)
            else:
                # back to compact form
                self.compact_label.setVisible(True)
                self.compact_status.setVisible(True)
                self.compact_show_popup.setIcon(
                    material_icon(
                        icon_name="expand_content", size=(10, 10), convert_to_pixmap=False
                    )
                )
                self.compact_view = True
                self.expand.emit(False)

    def setSizePolicy(self, size_policy1, size_policy2=None):
        # setting size policy on the compact popup widget will set
        # the policy for the container, and for itself
        if size_policy2 is None:
            # assuming first form: setSizePolicy(QSizePolicy)
            self.container.setSizePolicy(size_policy1)
            QWidget.setSizePolicy(self, size_policy1)
        else:
            self.container.setSizePolicy(size_policy1, size_policy2)
            QWidget.setSizePolicy(self, size_policy1, size_policy2)

    def addWidget(self, widget):
        """Add a widget to the popup container

        The popup container corresponds to the "full view" (not compact)
        The widget is reparented to the container, and added to the container layout
        """
        widget.setParent(self.container)
        self.container.layout().addWidget(widget)

    @Property(bool)
    def compact_view(self):
        return self._compact_view

    @compact_view.setter
    def compact_view(self, set_compact: bool):
        """Sets the compact form

        If set_compact is True, the compact view is displayed ; otherwise,
        the full view is displayed. This is handled by toggling visibility of
        the container widget or the compact view widget.
        """
        self._compact_view = set_compact
        if set_compact:
            self.compact_view_widget.setVisible(True)
            self.container.setVisible(False)
            QWidget.setSizePolicy(self, QSizePolicy.Fixed, QSizePolicy.Fixed)
        else:
            self.compact_view_widget.setVisible(False)
            self.container.setVisible(True)
            QWidget.setSizePolicy(self, self.container.sizePolicy())
        if self.parentWidget():
            self.parentWidget().adjustSize()
        else:
            self.adjustSize()

    @Property(str)
    def label(self):
        return self.compact_label.text()

    @label.setter
    def label(self, compact_label_text: str):
        """Set the label text associated to the compact view"""
        self.compact_label.setText(compact_label_text)

    @Property(str)
    def tooltip(self):
        return self.compact_label.toolTip()

    @tooltip.setter
    def tooltip(self, tooltip: str):
        """Set the tooltip text associated to the compact view"""
        self.compact_label.setToolTip(tooltip)
        self.compact_status.setToolTip(tooltip)

    @Property(bool)
    def expand_popup(self):
        return self._expand_popup

    @expand_popup.setter
    def expand_popup(self, popup: bool):
        self._expand_popup = popup



import bec_qthemes


def pretty_display_theme(theme: str = "dark"):
    palette = bec_qthemes.load_palette(theme)
    foreground = palette.text().color().name()
    background = palette.base().color().name()
    border = palette.shadow().color().name()
    accent = palette.accent().color().name()
    return f"""
QWidget {{color: {foreground}; background-color: {background}}}
QLabel {{ font-weight: bold; }}
QLineEdit,QLabel,QTreeView {{ border-style: solid; border-width: 2px; border-color: {border} }}
QRadioButton {{ color: {foreground}; }}
QRadioButton::indicator::checked {{ color: {accent}; }}
QCheckBox {{ color: {accent}; }}
"""


if __name__ == "__main__":
    print(pretty_display_theme())



"""Module for DapComboBox widget class to select a DAP model from a combobox."""

from bec_lib.logger import bec_logger
from qtpy.QtCore import Property, Signal, Slot
from qtpy.QtWidgets import QComboBox

from bec_widgets.utils.bec_widget import BECWidget

logger = bec_logger.logger


class DapComboBox(BECWidget, QComboBox):
    """
    Editable combobox listing the available DAP models.

    The widget behaves as a plain QComboBox and keeps ``fit_model_combobox`` as an alias to itself
    for backwards compatibility with older call sites.
    """

    ICON_NAME = "data_exploration"
    PLUGIN = True
    USER_ACCESS = ["select_y_axis", "select_x_axis", "select_fit_model"]

    ### Signals ###
    # Signal to emit a new dap_config: (x_axis, y_axis, fit_model). Can be used to add a new DAP process
    # in the BECWaveformWidget using its add_dap method. The signal is emitted when the user selects a new
    # fit model, but only if x_axis and y_axis are set.
    new_dap_config = Signal(str, str, str)
    # Signal to emit the name of the updated x_axis
    x_axis_updated = Signal(str)
    # Signal to emit the name of the updated y_axis
    y_axis_updated = Signal(str)
    # Signal to emit the name of the updated fit model
    fit_model_updated = Signal(str)

    def __init__(
        self,
        parent=None,
        client=None,
        gui_id: str | None = None,
        default_fit: str | None = None,
        **kwargs,
    ):
        super().__init__(parent=parent, client=client, gui_id=gui_id, **kwargs)
        self.fit_model_combobox = self  # Just for backwards compatibility with older call sites, the widget itself is the combobox
        self._available_models: list[str] = []
        self._x_axis = None
        self._y_axis = None
        self._is_valid_input = False

        self.setEditable(True)

        self.populate_fit_model_combobox()
        self.currentTextChanged.connect(self._on_text_changed)
        self.select_default_fit(default_fit)
        self.check_validity(self.currentText())

    def select_default_fit(self, default_fit: str | None = "GaussianModel"):
        """Set the default fit model.

        Args:
            default_fit(str): Default fit model.
        """
        if self._validate_dap_model(default_fit):
            self.select_fit_model(default_fit)
        elif self.available_models:
            self.select_fit_model(self.available_models[0])

    @property
    def available_models(self):
        """Available models property."""
        return self._available_models

    @available_models.setter
    def available_models(self, available_models: list[str]):
        """Set the available models.

        Args:
            available_models(list[str]): Available models.
        """
        self._available_models = available_models

    @Property(str)
    def x_axis(self):
        """X axis property."""
        return self._x_axis

    @x_axis.setter
    def x_axis(self, x_axis: str):
        """Set the x axis.

        Args:
            x_axis(str): X axis.
        """
        # TODO add validator for x axis -> Positioner? or also device (must be monitored)!!
        self._x_axis = x_axis
        self.x_axis_updated.emit(x_axis)

    @Property(str)
    def y_axis(self):
        """Y axis property."""
        # TODO add validator for y axis -> Positioner & Device? Must be a monitored device!!
        return self._y_axis

    @y_axis.setter
    def y_axis(self, y_axis: str):
        """Set the y axis.

        Args:
            y_axis(str): Y axis.
        """
        self._y_axis = y_axis
        self.y_axis_updated.emit(y_axis)

    @Slot(str)
    def _on_text_changed(self, fit_name: str):
        """
        Validate and emit updates for the current text.

        Args:
            fit_name(str): The current text in the combobox, representing the selected fit model.
        """
        self.check_validity(fit_name)
        if not self._is_valid_input:
            return

        self.fit_model_updated.emit(fit_name)
        if self.x_axis is not None and self.y_axis is not None:
            self.new_dap_config.emit(self._x_axis, self._y_axis, fit_name)

    @Slot(str)
    def check_validity(self, fit_name: str):
        """
        Highlight invalid manual entries similarly to DeviceComboBox.

        Args:
            fit_name(str): The current text in the combobox, representing the selected fit model.
        """
        if self._validate_dap_model(fit_name):
            self._is_valid_input = True
            self.setStyleSheet("border: 1px solid transparent;")
        else:
            self._is_valid_input = False
            if self.isEnabled():
                self.setStyleSheet("border: 1px solid red;")
            else:
                self.setStyleSheet("border: 1px solid transparent;")

    @Slot(str)
    def select_x_axis(self, x_axis: str):
        """Slot to update the x axis.

        Args:
            x_axis(str): X axis.
        """
        self.x_axis = x_axis
        self._on_text_changed(self.currentText())

    @Slot(str)
    def select_y_axis(self, y_axis: str):
        """Slot to update the y axis.

        Args:
            y_axis(str): Y axis.
        """
        self.y_axis = y_axis
        self._on_text_changed(self.currentText())

    @Slot(str)
    def select_fit_model(self, fit_name: str | None):
        """Slot to update the fit model.

        Args:
            fit_name(str): Fit model name.
        """
        if not self._validate_dap_model(fit_name):
            raise ValueError(f"Fit {fit_name} is not valid.")
        self.setCurrentText(fit_name)

    def populate_fit_model_combobox(self):
        """Populate the fit_model_combobox with the devices."""
        # pylint: disable=protected-access
        available_plugins = getattr(getattr(self.client, "dap", None), "_available_dap_plugins", {})
        self.available_models = [model for model in available_plugins.keys()]
        self.clear()
        self.addItems(self.available_models)

    def _validate_dap_model(self, model: str | None) -> bool:
        """Validate the DAP model.

        Args:
            model(str): Model name.
        """
        if model is None:
            return False
        return model in self.available_models

    @property
    def is_valid_input(self) -> bool:
        """Whether the current text matches an available DAP model."""
        return self._is_valid_input


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.utils.colors import apply_theme

    app = QApplication(sys.argv)
    apply_theme("dark")
    dialog = DapComboBox()
    dialog.show()
    sys.exit(app.exec_())



"""Module to create an arrow item for a pyqtgraph plot"""

import numpy as np
import pyqtgraph as pg
from bec_lib.logger import bec_logger
from qtpy.QtCore import QObject, QPointF, Signal, Slot

from bec_widgets.utils.colors import get_accent_colors

logger = bec_logger.logger


class BECIndicatorItem(QObject):

    def __init__(self, plot_item: pg.PlotItem = None, parent=None):
        super().__init__(parent=parent)
        self.accent_colors = get_accent_colors()
        self.plot_item = plot_item
        self._item_on_plot = False
        self._pos = None
        self.is_log_x = False
        self.is_log_y = False

    @property
    def item_on_plot(self) -> bool:
        """Returns if the item is on the plot"""
        return self._item_on_plot

    @item_on_plot.setter
    def item_on_plot(self, value: bool) -> None:
        self._item_on_plot = value

    def add_to_plot(self) -> None:
        """Add the item to the plot"""
        raise NotImplementedError("Method add_to_plot not implemented")

    def remove_from_plot(self) -> None:
        """Remove the item from the plot"""
        raise NotImplementedError("Method remove_from_plot not implemented")

    def set_position(self, pos) -> None:
        """This method should implement the logic to set the position of the
        item on the plot. Depending on the child class, the position can be
        a tuple (x,y) or a single value, i.e. x position where y position is fixed.
        """
        raise NotImplementedError("Method set_position not implemented")

    def check_log(self):
        """Checks if the x or y axis is in log scale and updates the internal state accordingly."""
        self.is_log_x = self.plot_item.ctrl.logXCheck.isChecked()
        self.is_log_y = self.plot_item.ctrl.logYCheck.isChecked()
        self.set_position(self._pos)


class BECTickItem(BECIndicatorItem):
    """Class to create a tick item which can be added to a pyqtgraph plot.
    The tick item will be added to the layout of the plot item and can be used to indicate
    a position"""

    position_changed = Signal(float)
    position_changed_str = Signal(str)

    def __init__(self, plot_item: pg.PlotItem = None, parent=None):
        super().__init__(plot_item=plot_item, parent=parent)
        self.tick_item = pg.TickSliderItem(
            parent=parent, allowAdd=False, allowRemove=False, orientation="bottom"
        )
        self.tick_item.skip_auto_range = True
        self.tick = None
        self._pos = 0.0
        self._range = [0, 1]

    @Slot(float)
    def set_position(self, pos: float) -> None:
        """Set the x position of the tick item

        Args:
            pos (float): The position of the tick item.
        """
        if self.is_log_x is True:
            pos = pos if pos > 0 else 1e-10
            pos = np.log10(pos)
        self._pos = pos
        view_box = self.plot_item.getViewBox()  # Ensure you're accessing the correct view box
        view_range = view_box.viewRange()[0]
        self.update_range(self.plot_item.vb, view_range)
        self.position_changed.emit(pos)
        self.position_changed_str.emit(str(pos))

    @Slot()
    def update_range(self, _, view_range: tuple[float, float]) -> None:
        """Update the range of the tick item

        Args:
            vb (pg.ViewBox): The view box.
            viewRange (tuple): The view range.
        """
        if self._pos < view_range[0] or self._pos > view_range[1]:
            self.tick_item.setVisible(False)
        else:
            self.tick_item.setVisible(True)

        if self.tick_item.isVisible():
            origin = self.tick_item.tickSize / 2.0
            length = self.tick_item.length

            length_with_padding = length + self.tick_item.tickSize + 2

            self._range = view_range
            tick_with_padding = (self._pos - view_range[0]) / (view_range[1] - view_range[0])
            tick_value = (tick_with_padding * length_with_padding - origin) / length
            self.tick_item.setTickValue(self.tick, tick_value)

    def add_to_plot(self):
        """Add the tick item to the view box or plot item."""
        if self.plot_item is None:
            return

        self.plot_item.layout.addItem(self.tick_item, 2, 1)
        self.tick_item.setOrientation("top")
        self.tick = self.tick_item.addTick(0, movable=False, color=self.accent_colors.highlight)
        self.update_tick_pos_y()
        self.plot_item.vb.sigXRangeChanged.connect(self.update_range)
        self.plot_item.ctrl.logXCheck.checkStateChanged.connect(self.check_log)
        self.plot_item.ctrl.logYCheck.checkStateChanged.connect(self.check_log)
        self.plot_item.vb.geometryChanged.connect(self.update_tick_pos_y)
        self.item_on_plot = True

    @Slot()
    def update_tick_pos_y(self):
        """Update tick position, while respecting the tick_item coordinates"""
        pos = self.tick.pos()
        pos = self.tick_item.mapToParent(pos)
        new_pos = self.plot_item.vb.geometry().bottom()
        new_pos = self.tick_item.mapFromParent(QPointF(pos.x(), new_pos))
        self.tick.setPos(new_pos)

    def remove_from_plot(self):
        """Remove the tick item from the view box or plot item."""
        if self.plot_item is not None and self.item_on_plot is True:
            self.plot_item.vb.sigXRangeChanged.disconnect(self.update_range)
            self.plot_item.ctrl.logXCheck.checkStateChanged.disconnect(self.check_log)
            self.plot_item.ctrl.logYCheck.checkStateChanged.disconnect(self.check_log)
            if self.plot_item.layout is not None:
                self.plot_item.layout.removeItem(self.tick_item)
        self.item_on_plot = False

    def cleanup(self) -> None:
        """Cleanup the item"""
        self.remove_from_plot()
        self.tick_item = None


class BECArrowItem(BECIndicatorItem):
    """Class to create an arrow item which can be added to a pyqtgraph plot.
    It can be either added directly to a view box or a plot item.
    To add the arrow item to a view box or plot item, use the add_to_plot method.

    Args:
        view_box (pg.ViewBox | pg.PlotItem): The view box or plot item to which the arrow item should be added.
        parent (QObject): The parent object.

    Signals:
        position_changed (tuple[float, float]): Signal emitted when the position of the arrow item has changed.
        position_changed_str (tuple[str, str]): Signal emitted when the position of the arrow item has changed.
    """

    # Signal to emit if the position of the arrow item has changed
    position_changed = Signal(tuple)
    position_changed_str = Signal(tuple)

    def __init__(self, plot_item: pg.PlotItem = None, parent=None):
        super().__init__(plot_item=plot_item, parent=parent)
        self.arrow_item = pg.ArrowItem()
        self.arrow_item.skip_auto_range = True
        self._pos = (0, 0)
        self.arrow_item.setVisible(False)

    @Slot(dict)
    def set_style(self, style: dict) -> None:
        """Set the style of the arrow item

        Args:
            style (dict): The style of the arrow item. Dictionary with key,
                          value pairs which are accepted from the pg.ArrowItem.setStyle method.
        """
        self.arrow_item.setStyle(**style)

    @Slot(tuple)
    def set_position(self, pos: tuple[float, float]) -> None:
        """Set the position of the arrow item

        Args:
            pos (tuple): The position of the arrow item as a tuple (x, y).
        """
        self._pos = pos
        pos_x = pos[0]
        pos_y = pos[1]
        if self.is_log_x is True:
            pos_x = np.log10(pos_x) if pos_x > 0 else 1e-10
            view_box = self.plot_item.getViewBox()  # Ensure you're accessing the correct view box
            view_range = view_box.viewRange()[0]
            # Avoid values outside the view range in the negative direction. Otherwise, there is
            # a buggy behaviour of the arrow item and it appears at the wrong position.
            if pos_x < view_range[0]:
                pos_x = view_range[0]
        if self.is_log_y is True:
            pos_y = np.log10(pos_y) if pos_y > 0 else 1e-10

        self.arrow_item.setPos(pos_x, pos_y)
        self.position_changed.emit(self._pos)
        self.position_changed_str.emit((str(self._pos[0]), str(self._pos[1])))

    def add_to_plot(self):
        """Add the arrow item to the view box or plot item."""
        if not self.arrow_item:
            logger.warning(f"Arrow item was already destroyed, cannot be created")
            return

        self.arrow_item.setStyle(
            angle=-90,
            pen=pg.mkPen(self.accent_colors.emergency, width=1),
            brush=pg.mkBrush(self.accent_colors.highlight),
            headLen=20,
        )
        self.arrow_item.setVisible(True)
        if self.plot_item is not None:
            self.plot_item.addItem(self.arrow_item)
            self.plot_item.ctrl.logXCheck.checkStateChanged.connect(self.check_log)
            self.plot_item.ctrl.logYCheck.checkStateChanged.connect(self.check_log)
            self.item_on_plot = True

    def remove_from_plot(self):
        """Remove the arrow item from the view box or plot item."""
        if self.plot_item is not None and self.item_on_plot is True:
            self.plot_item.ctrl.logXCheck.checkStateChanged.disconnect(self.check_log)
            self.plot_item.ctrl.logYCheck.checkStateChanged.disconnect(self.check_log)
            self.plot_item.removeItem(self.arrow_item)
        self.item_on_plot = False

    def cleanup(self) -> None:
        """Cleanup the item"""
        self.remove_from_plot()
        self.arrow_item = None
